/* Ankur Parikh
   CondensedSMat.cpp

   see .hpp for description
   */


#include "CondensedSMat.hpp"
#include "EigenHelper.hpp"
#include "SpecMoments.hpp"

using namespace Eigen;
using namespace std;


CondensedSMat::CondensedSMat()
{
    row_map = NULL;
    col_map = NULL;
    cond_mat = NULL;
}


CondensedSMat::~CondensedSMat()
{
    delete(row_map);
    delete(col_map);
    delete(cond_mat);
}


void CondensedSMat::Initialize(vector<triplet>& triplets, int V)
{
    row_map = new map<int, int>();
    col_map = new map<int, int>();

    map<int, double> col_sums;
    EigenHelper::ColSums(triplets, col_sums, V);
    map<int, double> row_sums;
    EigenHelper::RowSums(triplets, row_sums, V);

    int n_crows = 0;
    map<int, double>::iterator r_iter;
    for (r_iter = row_sums.begin(); r_iter != row_sums.end(); ++r_iter)
    {
        if (r_iter->second > 1e-10)
        {
            (*row_map)[r_iter->first] = n_crows++;
        }
    }

    int n_ccols = 0;
    map<int, double>::iterator c_iter;
    for (c_iter = col_sums.begin(); c_iter != col_sums.end(); ++c_iter)
    {
        if (c_iter->second > 1e-10)
        {
            (*col_map)[c_iter->first] = n_ccols++;
        }
    }

    cond_mat = new SparseMatrix<double>(n_crows, n_ccols);

    if (n_crows != 0 && n_ccols != 0)
    {
        vector<triplet> new_triplets;

        for (int i = 0; i < triplets.size(); i++)
        {
            triplet t = triplets[i];
            int r = t.row();
            int c = t.col();
            double val = t.value();
            assert(row_map->count(r) > 0 && col_map->count(c) > 0);
            new_triplets.push_back(triplet((*row_map)[r], (*col_map)[c], val));
        }

        cond_mat->setFromTriplets(new_triplets.begin(), new_triplets.end());
    }

}

bool CondensedSMat::ContainsRow(int i)
{
    return (row_map->count(i) > 0);
}

bool CondensedSMat::ContainsCol(int j)
{
    return (col_map->count(j) > 0);
}


CondensedSMat::CondensedSMat(SparseMatrix<double>& big_mat)
{
    row_map = new map<int, int>();
    col_map = new map<int, int>();

    VectorXd col_sums;
    EigenHelper::ColSums(big_mat, col_sums);
    VectorXd row_sums;
    EigenHelper::RowSums(big_mat, row_sums);

    int n_crows = 0;
    for (int i = 0; i < row_sums.size(); ++i)
    {
        if (row_sums(i) > 1e-10)
        {
            (*row_map)[i] = n_crows++;
        }
    }

    int n_ccols = 0;
    for (int i = 0; i < col_sums.size(); ++i)
    {
        if (col_sums(i) > 1e-10)
        {
            (*col_map)[i] = n_ccols++;
        }
    }



    cond_mat = new SparseMatrix<double>(n_crows, n_ccols);

    if (n_crows != 0 && n_ccols != 0)
    {
        vector<triplet> triplets;

        for (int i = 0; i < big_mat.outerSize(); i++)
        {
            for (SparseMatrix<double>::InnerIterator it(big_mat, i); it; ++it)
            {
                int r = it.row();
                int c = it.col();
                double val = it.value();

                if (val > 0)
                {
                    assert(row_map->count(r) > 0 && col_map->count(c) > 0);
                    triplets.push_back(triplet((*row_map)[r], (*col_map)[c], val));
                }
            }
        }

        cond_mat->setFromTriplets(triplets.begin(), triplets.end());
    }
}

void CondensedSMat::RightDiagMult(Eigen::SparseVector<double>& v)
{
    for (Eigen::SparseVector<double>::InnerIterator iter(v); iter; ++iter)
    {
        int r = iter.row();
        double val = iter.value();

        if (ContainsCol(r))
        {
            int c = (*col_map)[r];

            cond_mat->col(c) *= val;
        }
    }
}