/* Ankur Parikh
   EM.cpp

   see .hpp for class/function comments
   */


#include "EM.hpp"
#include <math.h>
#include <assert.h>
#include <limits>
#include <time.h>
#include "EigenHelper.hpp"

using namespace std;
using namespace Eigen;

void EM::RunBigParallelEM(MatrixXd& l_cpt, MatrixXd& r_cpt, MatrixXd& real_l_cpt,
    SparseMatrix<double>& counts, int K, int num_iters, int num_cores)
{

    srand(time(NULL));

    int n_rows = counts.rows();
    int n_cols = counts.cols();
    if (counts.sum() < 1e-6)
    {
        l_cpt.resize(n_rows, K);
        r_cpt.resize(K, n_cols);
        l_cpt.fill(0);
        r_cpt.fill(0);
        return;
    }


    VectorXd h_vec;
    MatrixXd cpt1;
    MatrixXd cpt2;

    // initialize parameters randomly
    GenerateRandomCPT(h_vec, K);
    GenerateRandomCPT(cpt1, n_rows, K);
    GenerateRandomCPT(cpt2, n_cols, K);

    double prev_ll = 0;


    VectorXd col_counts(counts.cols());
    col_counts.fill(0);
    for (int i = 0; i < counts.cols(); ++i)
    {
        col_counts(i) = counts.col(i).nonZeros();
    }

    // partition the counts for parallel processing
    vector<int> start_indices;
    vector<int> end_indices;
    EigenHelper::ParallelPartition(col_counts, start_indices, end_indices, num_cores);

    vector<VectorXd*> h_counts_vec;
    vector<MatrixXd*> x1h_counts_vec;
    vector<MatrixXd*> x2h_counts_vec;
    vector<double> ll_vec;

    // temporary core-specific agglomerate counts
    for (int c = 0; c < num_cores; ++c)
    {
        VectorXd* h_counts_c = new VectorXd(K);
        h_counts_c->fill(0);
        MatrixXd* x1h_counts_c = new MatrixXd(n_rows, K);
        x1h_counts_c->fill(0);
        MatrixXd* x2h_counts_c = new MatrixXd(n_cols, K);
        x2h_counts_c->fill(0);

        h_counts_vec.push_back(h_counts_c);
        x1h_counts_vec.push_back(x1h_counts_c);
        x2h_counts_vec.push_back(x2h_counts_c);
        ll_vec.push_back(0.0);
    }

    cout << "EM iteration: ";
    for (int n = 0; n < num_iters; ++n)
    {
        if (n % 5 == 0)
            cout << n << " ";
        cout.flush();
        VectorXd h_counts(K);
        h_counts.fill(0);
        MatrixXd x1h_counts(n_rows, K);
        x1h_counts.fill(0);
        MatrixXd x2h_counts(n_cols, K);
        x2h_counts.fill(0);

        map<pair<int, int>, double>::iterator iter;

        double ll = 0;



        // E-step
#pragma omp parallel for 
        for (int c = 0; c < num_cores; ++c)
        {
            int start_index = start_indices[c];
            int end_index = end_indices[c];

            for (int i = start_index; i < end_index; i++)
            {
                for (SparseMatrix<double>::InnerIterator it(counts, i); it; ++it)
                {
                    int x1 = it.row();
                    int x2 = it.col();
                    double count = it.value();

                    if (count == 0)
                        continue;

                    VectorXd x1_vec = cpt1.row(x1);
                    VectorXd x2_vec = cpt2.row(x2);

                    double p12 = x1_vec.transpose() * h_vec.asDiagonal() * x2_vec;
                    VectorXd h12vec = x1_vec.cwiseProduct(x2_vec).cwiseProduct(h_vec);

                    *(h_counts_vec[c]) += count * h12vec / p12;
                    (*(x1h_counts_vec[c])).row(x1) += count * h12vec / p12;
                    (*(x2h_counts_vec[c])).row(x2) += count * h12vec / p12;
                    ll_vec[c] += count * log(p12);

                    if (ll == numeric_limits<double>::infinity() || (-1 * ll) == numeric_limits<double>::infinity())
                    {
                        cout << "infinite: " << ll << "\t" << count << "\t" << p12 << "\t" << x1_vec.sum() << "\t" << x2_vec.sum() << "\t" << h_vec.sum() << "\t" << x2 << "\n";
                        cout << cpt2.row(x2) << "\n";
                        assert(0);
                    }

                }
            }
        }


        // agglomerate all the counts
        for (int c = 0; c < num_cores; ++c)
        {
            h_counts += *(h_counts_vec[c]);
            x1h_counts += *(x1h_counts_vec[c]);
            x2h_counts += *(x2h_counts_vec[c]);
            ll += ll_vec[c];
            h_counts_vec[c]->fill(0.0);
            x1h_counts_vec[c]->fill(0.0);
            x2h_counts_vec[c]->fill(0.0);
            ll_vec[c] = 0.0;
        }

        // convergence test based on likelihood
        if (n > 1)
        {
            double delta = abs((ll - prev_ll) / (abs(prev_ll) + pow(PREC, 2)));
            if (delta < PREC)
            {
                break;
            }
        }
        prev_ll = ll;

        // M step
        h_vec = h_counts / h_counts.sum();
        for (int k = 0; k < K; ++k)
        {
            cpt1.col(k) = x1h_counts.col(k) / (x1h_counts.col(k).sum());
            cpt2.col(k) = x2h_counts.col(k) / (x2h_counts.col(k).sum());
        }
    }

    for (int c = 0; c < num_cores; ++c)
    {
        delete(h_counts_vec[c]);
        delete(x1h_counts_vec[c]);
        delete(x2h_counts_vec[c]);
    }

    real_l_cpt = cpt1;
    l_cpt = cpt1 * h_vec.asDiagonal();
    r_cpt = cpt2.transpose();

    l_cpt = l_cpt * counts.sum();
    cout << "\n";
}

void EM::RunBigParallelEM(MatrixXd& l_cpt, MatrixXd& r_cpt,
    SparseMatrix<double>& counts, int K, int num_iters, int num_cores)
{

    srand(time(NULL));

    int n_rows = counts.rows();
    int n_cols = counts.cols();

    if (counts.sum() < 1e-6)
    {
        l_cpt.resize(n_rows, K);
        r_cpt.resize(K, n_cols);
        l_cpt.fill(0);
        r_cpt.fill(0);
        return;
    }


    VectorXd h_vec;
    MatrixXd cpt1;
    MatrixXd cpt2;

    // random initialization
    GenerateRandomCPT(h_vec, K);
    GenerateRandomCPT(cpt1, n_rows, K);
    GenerateRandomCPT(cpt2, n_cols, K);

    double prev_ll = 0;


    VectorXd col_counts(counts.cols());
    col_counts.fill(0);
    for (int i = 0; i < counts.cols(); ++i)
    {
        col_counts(i) = counts.col(i).nonZeros();
    }

    // partition the counts for parallel processing
    vector<int> start_indices;
    vector<int> end_indices;
    EigenHelper::ParallelPartition(col_counts, start_indices, end_indices, num_cores);

    vector<VectorXd*> h_counts_vec;
    vector<MatrixXd*> x1h_counts_vec;
    vector<MatrixXd*> x2h_counts_vec;
    vector<double> ll_vec;

    // temporary matrices to hold core-wise sufficient statistics
    for (int c = 0; c < num_cores; ++c)
    {
        VectorXd* h_counts_c = new VectorXd(K);
        h_counts_c->fill(0);
        MatrixXd* x1h_counts_c = new MatrixXd(n_rows, K);
        x1h_counts_c->fill(0);
        MatrixXd* x2h_counts_c = new MatrixXd(n_cols, K);
        x2h_counts_c->fill(0);

        h_counts_vec.push_back(h_counts_c);
        x1h_counts_vec.push_back(x1h_counts_c);
        x2h_counts_vec.push_back(x2h_counts_c);
        ll_vec.push_back(0.0);
    }

    cout << "EM iteration: ";
    for (int n = 0; n < num_iters; ++n)
    {
        cout << n << " ";
        cout.flush();
        VectorXd h_counts(K);
        h_counts.fill(0);
        MatrixXd x1h_counts(n_rows, K);
        x1h_counts.fill(0);
        MatrixXd x2h_counts(n_cols, K);
        x2h_counts.fill(0);

        map<pair<int, int>, double>::iterator iter;

        double ll = 0;


        // E step
#pragma omp parallel for 
        for (int c = 0; c < num_cores; ++c)
        {
            int start_index = start_indices[c];
            int end_index = end_indices[c];

            for (int i = start_index; i < end_index; i++)
            {
                for (SparseMatrix<double>::InnerIterator it(counts, i); it; ++it)
                {
                    int x1 = it.row();
                    int x2 = it.col();
                    double count = it.value();

                    if (count == 0)
                        continue;

                    VectorXd x1_vec = cpt1.row(x1);
                    VectorXd x2_vec = cpt2.row(x2);

                    double p12 = x1_vec.transpose() * h_vec.asDiagonal() * x2_vec;
                    VectorXd h12vec = x1_vec.cwiseProduct(x2_vec).cwiseProduct(h_vec);

                    *(h_counts_vec[c]) += count * h12vec / p12;
                    (*(x1h_counts_vec[c])).row(x1) += count * h12vec / p12;
                    (*(x2h_counts_vec[c])).row(x2) += count * h12vec / p12;
                    ll_vec[c] += count * log(p12);

                    if (ll == numeric_limits<double>::infinity() || (-1 * ll) == numeric_limits<double>::infinity())
                    {
                        cout << "infinite: " << ll << "\t" << count << "\t" << p12 << "\t" << x1_vec.sum() << "\t" << x2_vec.sum() << "\t" << h_vec.sum() << "\t" << x2 << "\n";
                        cout << cpt2.row(x2) << "\n";
                        assert(0);
                    }

                }
            }
        }


        // agglomerate all the counts
        for (int c = 0; c < num_cores; ++c)
        {
            h_counts += *(h_counts_vec[c]);
            x1h_counts += *(x1h_counts_vec[c]);
            x2h_counts += *(x2h_counts_vec[c]);
            ll += ll_vec[c];
            h_counts_vec[c]->fill(0.0);
            x1h_counts_vec[c]->fill(0.0);
            x2h_counts_vec[c]->fill(0.0);
            ll_vec[c] = 0.0;
        }

        // check convergence
        if (n > 1)
        {
            double delta = abs((ll - prev_ll) / (abs(prev_ll) + pow(PREC, 2)));
            if (delta < PREC)
            {
                break;
            }
        }
        prev_ll = ll;

        // M-step
        h_vec = h_counts / h_counts.sum();
        for (int k = 0; k < K; ++k)
        {
            cpt1.col(k) = x1h_counts.col(k) / (x1h_counts.col(k).sum());
            cpt2.col(k) = x2h_counts.col(k) / (x2h_counts.col(k).sum());
        }
    }

    for (int c = 0; c < num_cores; ++c)
    {
        delete(h_counts_vec[c]);
        delete(x1h_counts_vec[c]);
        delete(x2h_counts_vec[c]);
    }

    l_cpt = cpt1 * h_vec.asDiagonal();
    r_cpt = cpt2.transpose();

    l_cpt = l_cpt * counts.sum();
    cout << "\n";

}

void EM::RunBigEM(MatrixXd& l_cpt, MatrixXd& r_cpt,
    SparseMatrix<double>& counts, int K, int num_iters)
{
    srand(time(NULL));
    int n_rows = counts.rows();
    int n_cols = counts.cols();

    if (counts.sum() < 1e-6)
    {
        l_cpt.resize(n_rows, K);
        r_cpt.resize(K, n_cols);
        l_cpt.fill(0);
        r_cpt.fill(0);
        return;
    }


    VectorXd h_vec;
    MatrixXd cpt1;
    MatrixXd cpt2;


    // random initialization
    GenerateRandomCPT(h_vec, K);
    GenerateRandomCPT(cpt1, n_rows, K);
    GenerateRandomCPT(cpt2, n_cols, K);

    double prev_ll = 0;

    for (int n = 0; n < num_iters; ++n)
    {
        VectorXd h_counts(K);
        h_counts.fill(0);
        MatrixXd x1h_counts(n_rows, K);
        x1h_counts.fill(0);
        MatrixXd x2h_counts(n_cols, K);
        x2h_counts.fill(0);

        map<pair<int, int>, double>::iterator iter;

        double ll = 0;

        // E-step
        for (int i = 0; i < counts.outerSize(); i++)
        {
            for (SparseMatrix<double>::InnerIterator it(counts, i); it; ++it)
            {
                int x1 = it.row();
                int x2 = it.col();
                double count = it.value();

                if (count == 0)
                    continue;

                VectorXd x1_vec = cpt1.row(x1);
                VectorXd x2_vec = cpt2.row(x2);

                double p12 = x1_vec.transpose() * h_vec.asDiagonal() * x2_vec;
                VectorXd h12vec = x1_vec.cwiseProduct(x2_vec).cwiseProduct(h_vec);

                h_counts += count * h12vec / p12;
                x1h_counts.row(x1) += count * h12vec / p12;
                x2h_counts.row(x2) += count * h12vec / p12;


                ll += count * log(p12);

                if (ll == numeric_limits<double>::infinity() || (-1 * ll) == numeric_limits<double>::infinity())
                {
                    cout << "infinite: " << ll << "\t" << count << "\t" << p12 << "\t" << x1_vec.sum() << "\t" << x2_vec.sum() << "\t" << h_vec.sum() << "\t" << x2 << "\n";
                    cout << cpt2.row(x2) << "\n";
                    assert(0);
                }
            }
        }

        // check convergence
        if (n > 1)
        {
            double delta = abs((ll - prev_ll) / (abs(prev_ll) + pow(PREC, 2)));
            if (delta < PREC)
            {
                break;
            }
        }
        prev_ll = ll;

        // M step
        h_vec = h_counts / h_counts.sum();
        for (int k = 0; k < K; ++k)
        {
            cpt1.col(k) = x1h_counts.col(k) / (x1h_counts.col(k).sum());
            cpt2.col(k) = x2h_counts.col(k) / (x2h_counts.col(k).sum());
        }
    }
    l_cpt = cpt1 * h_vec.asDiagonal();
    r_cpt = cpt2.transpose();

    l_cpt = l_cpt * counts.sum();
}

void EM::GenerateRandomCPT(VectorXd& v, int K)
{
    v.resize(K);

    for (int i = 0; i < K; ++i)
    {
        v(i) = rand() + 1e-10;
    }

    v = v / v.sum();
}

void EM::GenerateRandomCPT(MatrixXd& m, int V, int K)
{
    m.resize(V, K);

    for (int j = 0; j < K; ++j)
    {
        for (int i = 0; i < V; ++i)
        {
            m(i, j) = rand() + 1e-10;
        }

        m.col(j) = m.col(j) / m.col(j).sum();
    }
}
