/* Ankur Parikh
   EigenHelper.hpp

   see .hpp for class/function descriptions
   */

#include "EigenHelper.hpp"
#include <iostream>
#include "SpecMoments.hpp"

using namespace std;


MatrixXd EigenHelper::pinv(MatrixXd& m1)
{
    JacobiSVD<MatrixXd> svd(m1, ComputeThinU | ComputeThinV);
    MatrixXd U = svd.matrixU();
    MatrixXd V = svd.matrixV();
    VectorXd s_vals = svd.singularValues();

    MatrixXd Sigma(s_vals.size(), s_vals.size());
    Sigma.fill(0);

    for (int s = 0; s < s_vals.size(); ++s)
    {
        if (abs(s_vals(s)) < .00000000001)
            assert(0);
        Sigma(s, s) = 1.0 / s_vals(s);

    }

    MatrixXd pinverse_temp = V * Sigma;
    return  pinverse_temp * U.transpose();
}



void EigenHelper::PrintVec(VectorXd& v1)
{
    for (int i = 0; i < v1.size(); ++i)
    {
        cout << v1(i);
        cout << "\n";
    }
}

void EigenHelper::PrintMat(MatrixXd& m1)
{
    for (int i = 0; i < m1.rows(); ++i)
    {
        for (int j = 0; j < m1.cols(); ++j)
        {
            cout << m1(i, j);
            cout << "\t";
        }
        cout << "\n";
    }
}

void EigenHelper::PrintSpMat(SparseMatrix<double> spmat){
    for (int i = 0; i < spmat.outerSize(); i++){
        for (SparseMatrix<double>::InnerIterator it(spmat, i); it; ++it){
            cout << it.row() << "\t" << it.col() << "\t" << it.value() << "\n";
        }
    }
}

void EigenHelper::ConvertToBinary(SparseMatrix<double>& spmat, SparseMatrix<double>& b_spmat)
{
    vector<triplet> triplet_vec;
    for (int i = 0; i < spmat.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(spmat, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double val = it.value();

            triplet_vec.push_back(triplet(r, c, 1.0));
        }
    }
    b_spmat.resize(spmat.rows(), spmat.cols());
    b_spmat.setFromTriplets(triplet_vec.begin(), triplet_vec.end());
}


double EigenHelper::Min(SparseMatrix<double>& m)
{
    double curr_min = 1000000;
    for (int i = 0; i < m.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(m, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double val = it.value();
            if (val < curr_min)
                curr_min = val;
        }
    }
    assert(curr_min < 1000000);
    return curr_min;
}


void EigenHelper::Power(SparseMatrix<double>& m, SparseMatrix<double>& power_m, double P)
{
    vector<triplet> triplet_vec;
    for (int i = 0; i < m.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(m, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double val = it.value();
            double power_val = pow(val, P);
            if (val == 0.0)
                power_val = 0.0;
            triplet_vec.push_back(triplet(r, c, power_val));
        }
    }
    power_m.resize(m.rows(), m.cols());
    power_m.setFromTriplets(triplet_vec.begin(), triplet_vec.end());
}




void EigenHelper::RowSums(SparseMatrix<double>& m, VectorXd& rowsums)
{
    rowsums.resize(m.rows());
    rowsums.fill(0);
    vector<triplet> triplets;

    for (int i = 0; i < m.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(m, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double val = it.value();

            rowsums(r) += val;
        }
    }
}

void EigenHelper::ColSums(vector< Triplet<double> >& triplets, map<int, double>& colsums, int V)
{
    double total_sum = 0;
    for (int i = 0; i < triplets.size(); i++)
    {
        triplet t = triplets[i];
        int r = t.row();
        int c = t.col();
        double val = t.value();

        if (colsums.count(c) == 0)
            colsums[c] = 0;

        colsums[c] += val;
        total_sum += val;
    }
}

void EigenHelper::RowSums(vector< Triplet<double> >& triplets, map<int, double>& rowsums, int V)
{
    for (int i = 0; i < triplets.size(); i++)
    {
        triplet t = triplets[i];
        int r = t.row();
        int c = t.col();
        double val = t.value();

        if (rowsums.count(r) == 0)
            rowsums[r] = 0;

        rowsums[r] += val;
    }
}

void EigenHelper::ColSums(SparseMatrix<double>& m, VectorXd& colsums)
{
    colsums.resize(m.cols());
    colsums.fill(0);

    for (int i = 0; i < m.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(m, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            double val = it.value();

            colsums(c) += val;
        }
    }
}

void EigenHelper::ColSums(SparseMatrix<int>& m, VectorXd& colsums)
{
    colsums.resize(m.cols());
    colsums.fill(0);

    for (int i = 0; i < m.outerSize(); i++)
    {
        for (SparseMatrix<int>::InnerIterator it(m, i); it; ++it)
        {
            int r = it.row();
            int c = it.col();
            int val = it.value();

            colsums(c) += val;
        }
    }
}



void EigenHelper::DenseInverse(VectorXd& p_denom, VectorXd& p_denom_inv)
{
    p_denom_inv = p_denom;
    for (int i = 0; i < p_denom.size(); ++i)
    {
        if (p_denom(i) != 0.0)
            p_denom_inv(i) = 1.0 / p_denom(i);
    }
}

void EigenHelper::SparseInverse(Eigen::SparseVector<double>& v, Eigen::SparseVector<double>& inv_v)
{
    inv_v = v;
    for (Eigen::SparseVector<double>::InnerIterator iter(v); iter; ++iter)
    {
        int r = iter.row();
        double val = iter.value();
        if (val != 0.0)
        {
            inv_v.coeffRef(r) = 1.0 / val;
        }
    }
}

void EigenHelper::ParallelPartition(VectorXd& counts, vector<int>& start_indices, vector<int>& end_indices, int num_cores)
{
    int total_nnz = counts.sum();
    int curr_index = 0;
    int p_size = total_nnz / num_cores;
    for (int i = 0; i < num_cores; ++i)
    {
        int curr_size = 0;
        if (i == 0)
        {
            start_indices.push_back(0);
        }
        else
        {
            start_indices.push_back(curr_index);
            assert(curr_index < counts.size());
        }

        if (i == num_cores - 1)
        {
            end_indices.push_back(counts.size());
        }
        else
        {
            while (curr_size < p_size)
            {
                curr_size += counts(curr_index);
                curr_index++;
            }
            end_indices.push_back(curr_index);
        }
    }
}