/* Ankur Parikh
   IKNWrapper.cpp

   see .hpp for class/function comments
   */

#include "IKNWrapper.hpp"
#include <assert.h>

using namespace std;
using namespace Eigen;


IKNWrapper::IKNWrapper(SpecMoments* spec_mom, IKneserNey* kn,
    IQuadKneserNey* qkn, int max_order)
{
    this->spec_mom = spec_mom;
    this->kn = kn;
    this->qkn = qkn;

    this->max_order = max_order;

    quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
    triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
    pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
}


IKNWrapper::IKNWrapper(SpecMoments* spec_mom, IKneserNey* kn, int max_order)
{
    this->spec_mom = spec_mom;
    this->kn = kn;
    this->qkn = NULL;
    this->max_order = max_order;

    quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
    triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
    pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
}



IKNWrapper::~IKNWrapper()
{
    delete(kn);
    if (qkn != NULL)
        delete(qkn);
    delete(quad_cache);
    delete(triple_cache);
    delete(pair_cache);
}

double IKNWrapper::GetCondProb(const string& w, const string& wp,
    const string& wpp, const string& wppp)
{
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    int k = spec_mom->GetFinalID(wpp);
    int l = spec_mom->GetFinalID(wppp);

    iquadple q(i, j, k, l);
    if (quad_cache->count(q) > 0)
        return (*quad_cache)[q];
    else
    {
        double val = GetCondProb(i, j, k, l);
        (*quad_cache)[q] = val;
        if (quad_cache->size() > 10000)
        {
            cerr << "cache cleared\n";
            quad_cache->clear();
        }
        return val;
    }


}

double IKNWrapper::GetCondProb(const string& w, const string& wp, const string& wpp)
{
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    int k = spec_mom->GetFinalID(wpp);
    itriple t(i, j, k);
    if (triple_cache->count(t) > 0)
        return (*triple_cache)[t];
    else
    {
        double val = GetCondProb(i, j, k);
        (*triple_cache)[t] = val;
        if (triple_cache->size() > 10000)
        {
            cerr << "cache cleared\n";
            triple_cache->clear();
        }
        return val;
    }
}

double IKNWrapper::GetCondProb(const string& w, const string& wp)
{
    int i = spec_mom->GetFinalID(w);
    int j = spec_mom->GetFinalID(wp);
    pair<int, int> p(i, j);
    if (pair_cache->count(p) > 0)
        return (*pair_cache)[p];
    else
    {
        double val = GetCondProb(i, j);
        (*pair_cache)[p] = val;
        if (pair_cache->size() > 10000)
        {
            cerr << "cache cleared\n";
            pair_cache->clear();
        }

        return val;
    }
}

double IKNWrapper::GetCondProb(const string& w)
{
    //	assert(wp == start_marker);
    int i = spec_mom->GetFinalID(w);
    return GetCondProb(i);
}


double IKNWrapper::GetCondProb(int i, int j, int k, int l)
{
    if (qkn != NULL && max_order >= 4)
    {
        double prob = qkn->GetProb(i, j, k, l);
        return prob;
    }
    else
    {
        return GetCondProb(i, j, k);
    }
}

double IKNWrapper::GetCondProb(int i, int j, int k)
{
    if (max_order >= 3)
    {
        double kn_bweight = kn->GetKneserWeight(j, k);
        double kn_uweight = kn->GetKneserWeight(j);

        double kn_tri_prob = kn->GetDiscountedTrigramProb(i, j, k);
        double kn_bi_prob = kn->GetDiscountedBigramProb(i, j);
        double kn_uni_prob = kn->GetDiscountedUnigramProb(i);
        double total_prob = kn_tri_prob + (kn_bweight)*(kn_bi_prob + kn_uweight * kn_uni_prob);

        return total_prob;
    }
    else
    {
        return GetCondProb(i, j);
    }
}

double IKNWrapper::GetCondProb(int i, int j)
{
    assert(max_order >= 2);
    double u_weight = kn->GetBiKneserWeight(j);

    double mle_bi_prob = kn->GetBiDiscBigramProb(i, j);
    double unigram_prob = kn->GetDiscountedUnigramProb(i);
    double total_prob = mle_bi_prob + u_weight * unigram_prob;

    return total_prob;
}

double IKNWrapper::GetCondProb(int i)
{
    VectorXd& unigrams = spec_mom->getUnigrams();
    double prob = unigrams(i);
    return prob;
}