/* Ankur Parikh
IKneserNey.hpp

Interpolated 3rd and 2nd order Kneser Ney Models

TODO: Refactor 2,3,4th order models into a single framework
*/


#pragma once 

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include <limits>
#include <vector>
#include <set>
#include "eigen_boost_serialization.hpp"

using namespace std;
using namespace Eigen;


class IKneserNey
{

private:

    SpecMoments* spec_mom; // moments
    IWittenBell* wb; // witten bell
    bool m_flag; // true if modified version, false if standard

    // counts of counts for lower order distributions
    map<pair<int, int>, int>* b_n1prev_counts; // bigrams
    SparseMatrix<double>* b_n1prev_mat; // bigrams in matrix form
    map<int, int>* u_n1prev_counts; // unigrams

    VectorXd* denom_b_n1prev; // normalizer for bigrams
    int denom_u_n1prev; // normalizer for unigrams

    // just for testing, can be ignored
    VectorXd* denom_b_n1test;
    double denom_n1test_sum;

    // discounts for third order model (trigram)
    double D1;
    double D2;
    double D3plus;
    double max_d_limit;

    // discounts for third order model (bigram)
    double bD1;
    double bD2;
    double bD3plus;
    double b_max_d_limit;

    // discount for 2nd order model
    double bi_D1;
    double bi_D2;
    double bi_D3plus;
    double bi_max_d_limit;

    // unigram discount
    double uni_D1;

    // gammas for 3rd order model
    map<pair<int, int>, double>* bidisc_map;
    map<int, double>* unidisc_map;

    // gamma for 2nd order model
    map<int, double>* bi_unidisc_map;

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & spec_mom;
        ar & wb;
        ar & m_flag;

        ar & b_n1prev_counts;
        ar & b_n1prev_mat;
        ar & u_n1prev_counts;

        ar & denom_b_n1prev;
        ar & denom_u_n1prev;

        ar & denom_b_n1test;
        ar & denom_n1test_sum;

        ar & D1;
        ar & D2;
        ar & D3plus;
        ar & max_d_limit;

        ar & bD1;
        ar & bD2;
        ar & bD3plus;
        ar & b_max_d_limit;

        ar & bi_D1;
        ar & bi_D2;
        ar & bi_D3plus;
        ar & bi_max_d_limit;

        ar & bidisc_map;
        ar & unidisc_map;
        ar & bi_unidisc_map;
    }

public:

    IKneserNey()
    {
        b_n1prev_counts = NULL;
        u_n1prev_counts = NULL;
        denom_b_n1prev = NULL;
        denom_u_n1prev = 0;

        D1 = -1;
        D2 = -1;
        D3plus = -1;
        max_d_limit = -1;

        bi_D1 = -1;
        bi_D2 = -1;
        bi_D3plus = -1;
        bi_max_d_limit = -1;

        bD1 = -1;
        bD2 = -1;
        bD3plus = -1;
        b_max_d_limit = -1;
    }

    IKneserNey(SpecMoments* spec_mom, IWittenBell* wb, bool m_flag);

    IKneserNey(SpecMoments* spec_mom, IWittenBell* wb, double D1, double D2, double D3plus,
        double bD1, double bD2, double bD3plus);

    ~IKneserNey()
    {
        delete(wb);
        delete(b_n1prev_counts);
        delete(b_n1prev_mat);
        delete(u_n1prev_counts);
        delete(denom_b_n1prev);
        delete(denom_b_n1test);
        delete(bidisc_map);
        delete(unidisc_map);
        delete(bi_unidisc_map);
    }

    // compute counts of counts
    void ComputeUCounts();
    void ComputeBCounts();

    void ComputeDTrigrams();

    // compute leftover weights
    void ComputeBWeights();
    void ComputeUWeights();

    double GetNonDiscBigramProb(int i, int j);

    void ComputeKneserDiscounts();

    // discount accessors (third order) for trigram
    double GetTriD1() { return D1; }
    double GetTriD2() { return D2; }
    double GetTriD3() { return D3plus; }

    // discount accessors (third order) for bigram
    double GetBD1() { return bD1; }
    double GetBD2() { return bD2; }
    double GetBD3() { return bD3plus; }

    // discount accessors (second order) for bigram
    double GetBiD1() { return bi_D1; }
    double GetBiD2() { return bi_D2; }
    double GetBiD3() { return bi_D3plus; }

    double GetUniD1() { return uni_D1; }


    double GetDiscountedTrigramProb(int i, int j, int k);
    double GetDiscountedBigramProb(int i, int j);
    double GetDiscountedUnigramProb(int i);
    double GetTriDiscountedUnigramProb(int i);

    SparseMatrix<double>& GetBN1PrevMat()
    {
        return *b_n1prev_mat;
    }

    void ComputeTriUniKneserDiscounts();
    void ComputeBiKneserDiscounts();
    void ComputeBiUWeights();
    double GetBiDiscBigramProb(int i, int j);

    int GetTotalN1Prev()
    {
        return (int)(b_n1prev_mat->sum());
    }

    double GetTotalN1Prev(int i)
    {
        return (denom_b_n1prev->coeff(i));
    }

    // 3rd order leftover weight
    double GetKneserWeight(int i, int j)
    {
        pair<int, int> p(i, j);

        if (bidisc_map->count(p) > 0)
        {
            double val = (*bidisc_map)[p];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }
    }

    // 3rd order leftover weight
    double GetKneserWeight(int i)
    {
        if (unidisc_map->count(i) > 0)
        {
            double val = (*unidisc_map)[i];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }

    }

    // 2nd order model leftover weight
    double GetBiKneserWeight(int i)
    {
        if (bi_unidisc_map->count(i) > 0)
        {
            double val = (*bi_unidisc_map)[i];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }
    }


    double GetN1Prev(int i, int j);
    double GetN1Prev(int i);



};
