/* Ankur Parikh
    IQuadKneserNey.hpp

    implements 4th order Kneser Ney (modified interpolated)
    */

#pragma once 

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include <limits>
#include <vector>
#include <set>
#include "eigen_boost_serialization.hpp"

using namespace std;
using namespace Eigen;


class IQuadKneserNey
{

private:

    SpecMoments* spec_mom; // moments
    IWittenBell* wb; // witten bell
    bool m_flag; // true if modified kn, false if normal

    map<itriple, int>* t_n1prev_counts; // lower order altered trigram counts
    map<pair<int, int>, int>* b_n1prev_counts; // lower order bigram counts
    SparseMatrix<double>* b_n1prev_mat; // lower order bigram counts in matrix form
    map<int, int>* u_n1prev_counts; // lower order unigram counts

    // MatrixXd* denom_t_n1prev;
    SparseMatrix<double>* denom_t_n1prev; // lower order altered trigram denominator
    VectorXd* denom_b_n1prev; // lower order altered bigram denominator
    int denom_u_n1prev; // lower order altered unigram denominator

    // just for testing can be ignored
    VectorXd* denom_b_n1test;
    double denom_n1test_sum;

    // quadgram discounts
    double qD1;
    double qD2;
    double qD3plus;
    double q_max_d_limit;

    // trigram discounts
    double tD1;
    double tD2;
    double tD3plus;
    double t_max_d_limit;

    // bigram discounts
    double bD1;
    double bD2;
    double bD3plus;
    double b_max_d_limit;

    // leftover weights
    map<itriple, double>* tridisc_map;
    map<pair<int, int>, double>* bidisc_map;
    map<int, double>* unidisc_map;
    // map<int, double>* bi_unidisc_map;

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {
        ar & denom_t_n1prev;
        //  cout << "yee1\n";
        ar & spec_mom;
        ar & wb;
        ar & m_flag;
        ar & t_n1prev_counts;
        ar & b_n1prev_counts;
        ar & b_n1prev_mat;
        ar & u_n1prev_counts;

        ar & denom_b_n1prev;
        ar & denom_u_n1prev;
        ar & denom_b_n1test;
        ar & denom_n1test_sum;
        ar & qD1;
        ar & qD2;
        ar & qD3plus;
        ar & q_max_d_limit;

        ar & tD1;
        ar & tD2;
        ar & tD3plus;
        ar & t_max_d_limit;

        ar & bD1;
        ar & bD2;
        ar & bD3plus;
        ar & b_max_d_limit;
        ar & tridisc_map;
        ar & bidisc_map;
        ar & unidisc_map;
        //	ar & bi_unidisc_map;
    }

public:

    IQuadKneserNey()
    {
        b_n1prev_counts = NULL;
        u_n1prev_counts = NULL;
        denom_b_n1prev = NULL;
        denom_u_n1prev = 0;

        tD1 = -1;
        tD2 = -1;
        tD3plus = -1;
        t_max_d_limit = -1;

        bD1 = -1;
        bD2 = -1;
        bD3plus = -1;
        b_max_d_limit = -1;
    }

    IQuadKneserNey(SpecMoments* spec_mom, IWittenBell* wb, bool m_flag);

    IQuadKneserNey(SpecMoments* spec_mom, IWittenBell* wb, double D1, double D2, double D3plus,
        double bD1, double bD2, double bD3plus);

    ~IQuadKneserNey()
    {
        delete(wb);
        delete(t_n1prev_counts);
        delete(b_n1prev_counts);
        delete(b_n1prev_mat);
        delete(u_n1prev_counts);
        delete(denom_t_n1prev);
        delete(denom_b_n1prev);
        delete(denom_b_n1test);
        delete(tridisc_map);
        delete(bidisc_map);
        delete(unidisc_map);
    }

    double GetProb(int i, int j, int k, int l);
    double GetQTProb(int i, int j, int k, int l);
    double GetQTBProb(int i, int j, int k, int l);

    // compute alternate lower order distributions
    void ComputeUCounts(); // unigram
    void ComputeBCounts(); // bigram
    void ComputeTCounts(); // trigram

    // compute leftover weights
    void ComputeTWeights(); //trigram
    void ComputeBWeights(); // bigram
    void ComputeUWeights();//unigram 

    // compute discounts using good turing estimation
    void ComputeQuadKneserDiscounts();
    void ComputeTriKneserDiscounts();
    void ComputeBiKneserDiscounts();
    void ComputeTriUniKneserDiscounts();

    // discount accessors
    double GetQD1() { return qD1; }
    double GetQD2() { return qD2; }
    double GetQD3() { return qD3plus; }

    double GetTriD1() { return tD1; }
    double GetTriD2() { return tD2; }
    double GetTriD3() { return tD3plus; }

    double GetBD1() { return bD1; }
    double GetBD2() { return bD2; }
    double GetBD3() { return bD3plus; }

    // discounted probability accessors
    double GetDiscountedQuadgramProb(int i, int j, int k, int l);

    double GetDiscountedTrigramProb(int i, int j, int k);
    double GetNonDiscTrigramProb(int i, int j, int k);

    double GetDiscountedBigramProb(int i, int j);
    double GetNonDiscBigramProb(int i, int j);

    double GetDiscountedUnigramProb(int i);

    double GetTriDiscountedUnigramProb(int i);


    double GetBiDiscBigramProb(int i, int j);

    int GetTotalN1Prev()
    {
        return (int)(b_n1prev_mat->sum());
    }

    int GetTotalN1Prev(int i)
    {
        return (denom_b_n1prev->coeff(i));
    }
    // trigram weight
    double GetKneserWeight(int i, int j, int k)
    {
        itriple t(i, j, k);

        if (tridisc_map->count(t) > 0)
        {
            double val = (*tridisc_map)[t];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }
    }
    // bigram weight
    double GetKneserWeight(int i, int j)
    {
        pair<int, int> p(i, j);

        if (bidisc_map->count(p) > 0)
        {
            double val = (*bidisc_map)[p];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }
    }

    // unigram  weight
    double GetKneserWeight(int i)
    {
        if (unidisc_map->count(i) > 0)
        {
            double val = (*unidisc_map)[i];
            assert(val > 0);
            return val;
        }
        else
        {
            return 1;
        }

    }

    // count of count accessors
    double GetN1Prev(int i, int j, int k);
    double GetN1Prev(int i, int j);
    double GetN1Prev(int i);
};
