/* Ankur Parikh
  PLRE.cpp

  learns 3rd order and 2nd order PLRE models */

#include "PLRE.hpp"
#include "IOLibrary.hpp"
#include "DiscountComputer.hpp"
#include "EM.hpp"
#include "PowerFullTrigram.hpp"
#include "PowerFullBigram.hpp"
#include "EigenHelper.hpp"

using namespace std;
using namespace Eigen;



PLRE::PLRE(SpecMoments* spec_mom, vector<double> tri_fvec, vector<double> bi_fvec,
    vector<double> tri_pvec, vector<double> bi_pvec, vector<double> tri_dvec, vector<double> bi_dvec, double uni_D)
{
    this->tri_fvec = new vector<double>(tri_fvec);
    this->bi_fvec = new vector<double>(bi_fvec);
    this->tri_pvec = new vector<double>(tri_pvec);
    this->bi_pvec = new vector<double>(bi_pvec);
    this->tri_dvec = new vector<double>(tri_dvec);
    this->bi_dvec = new vector<double>(bi_dvec);
    this->V = spec_mom->GetV();
    this->spec_mom = spec_mom;

    this->bigram_vec = new vector<PowerBigram*>();
    this->trigram_vec = new vector<PowerTrigram*>();


    SMat3Tensor<int>& trigram_counts = spec_mom->getTrigramCMat();
    SparseMatrix<double>* trigram_cmat = new SparseMatrix<double>(trigram_counts.GetMat().cast<double>());

    vector< SparseMatrix<double>* > p_trigrams;
    vector< SparseMatrix<double>* > tri_disc_mats;
    vector < VectorXd* > tri_ungammas;

    // compute all the powers and discounts for the trigrams
    p_trigrams.push_back(trigram_cmat);
    for (int i = 0; i < tri_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_trigram = new SparseMatrix<double>();
        SparseMatrix<double>* tri_disc_mat = new SparseMatrix<double>();
        VectorXd* tri_ungamma = new VectorXd();

        EigenHelper::Power(*trigram_cmat, *p_trigram, tri_pvec[i]);
        double prev_P = 1.0;
        if (i > 0)
            prev_P = tri_pvec[i - 1];

        DiscountComputer::ComputeDiscount(*tri_disc_mat, *tri_ungamma, *(p_trigrams[i]), *p_trigram, prev_P, tri_dvec);

        p_trigrams.push_back(p_trigram);
        tri_disc_mats.push_back(tri_disc_mat);
        tri_ungammas.push_back(tri_ungamma);
    }
    assert(tri_pvec[tri_pvec.size() - 1] == 0);


    VectorXd curr_gamma;


    // compute the low rank approximations of these powered discount trigrams
    for (int i = 0; i < p_trigrams.size() - 1; ++i)
    {
        if (i == 0)
        {
            PowerTrigram* lr_trigram = new PowerFullTrigram(&trigram_counts, *(tri_disc_mats[i]), *(tri_ungammas[i]));
            trigram_vec->push_back(lr_trigram);
            curr_gamma = *(tri_ungammas[i]);
        }
        else
        {
            VectorXd p_denom;
            VectorXd p_denom_inv;
            EigenHelper::ColSums(*(p_trigrams[i]), p_denom);
            EigenHelper::DenseInverse(p_denom, p_denom_inv);

            cout << "start computing low rank trigram\n";
            PowerLRTrigram* lr_trigram = new PowerLRTrigram(&trigram_counts, *(p_trigrams[i]), *(tri_disc_mats[i]),
                curr_gamma, *(tri_ungammas[i]), -1, tri_fvec[i - 1]);
            cout << "done computing low rank trigram\n";

            trigram_vec->push_back(lr_trigram);

            ArrayXd temp_gamma = (*(tri_ungammas[i])).array() * p_denom_inv.array();
            temp_gamma = curr_gamma.array() * temp_gamma;
            curr_gamma = temp_gamma.matrix();
            curr_gamma = *(tri_ungammas[i]);
        }

    }

    // create the bigrams (rank 1, power 0 trigram)
    SparseMatrix<double>* b_counts = new SparseMatrix<double>();
    ComputeBCounts(*b_counts, trigram_counts, *(p_trigrams[p_trigrams.size() - 1]));

    vector< SparseMatrix<double>* > p_bigrams;
    vector< SparseMatrix<double>* > bi_disc_mats;
    vector < VectorXd* > bi_ungammas;

    // compute powered bigrams
    p_bigrams.push_back(b_counts);
    for (int i = 0; i < tri_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_bigram = new SparseMatrix<double>();
        SparseMatrix<double>* bi_disc_mat = new SparseMatrix<double>();
        VectorXd* bi_ungamma = new VectorXd();

        EigenHelper::Power(*b_counts, *p_bigram, bi_pvec[i]);

        double prev_P = 1.0;
        if (i > 0)
            prev_P = bi_pvec[i - 1];

        DiscountComputer::ComputeDiscount(*bi_disc_mat, *bi_ungamma, *(p_bigrams[i]), *p_bigram, prev_P, bi_dvec);

        p_bigrams.push_back(p_bigram);
        bi_disc_mats.push_back(bi_disc_mat);
        bi_ungammas.push_back(bi_ungamma);
    }

    SparseMatrix<double>* zero_bdisc = new SparseMatrix<double>(V, V);
    VectorXd* zero_gamma = new VectorXd(V);
    zero_gamma->fill(0);

    bi_disc_mats.push_back(zero_bdisc);
    bi_ungammas.push_back(zero_gamma);

    // take low rank approximations
    VectorXd b_curr_gamma;
    for (int i = 0; i < p_bigrams.size(); ++i)
    {
        if (i == 0)
        {
            PowerBigram* lr_bigram = new PowerFullBigram(*(p_bigrams[i]), *(bi_disc_mats[i]), *(bi_ungammas[i]));
            bigram_vec->push_back(lr_bigram);
            b_curr_gamma = *(bi_ungammas[i]);
        }
        else
        {
            cout << "computing low rank bigram\n";
            PowerBigram* lr_bigram = new PowerLRBigram(*(p_bigrams[i]), *(bi_disc_mats[i]), b_curr_gamma,
                *(bi_ungammas[i]), bi_fvec[i - 1]);
            cout << "done computing low rank bigram\n";
            bigram_vec->push_back(lr_bigram);

            VectorXd p_denom;
            VectorXd p_denom_inv;
            EigenHelper::ColSums(*(p_bigrams[i]), p_denom);
            EigenHelper::DenseInverse(p_denom, p_denom_inv);

            ArrayXd temp_gamma = (*(bi_ungammas[i])).array() * p_denom_inv.array();
            temp_gamma = b_curr_gamma.array() * temp_gamma;
            b_curr_gamma = temp_gamma.matrix();
            b_curr_gamma = *(bi_ungammas[i]);

        }
    }

    // unigram counts (power 0, rank 1 bigram)
    VectorXd u_counts;
    ComputeUCounts(u_counts, *(p_bigrams[p_bigrams.size() - 1]));
    ComputeSmoothedUnigram(u_counts, uni_D);

    // delete intermediate quantities

    for (int i = 0; i < p_trigrams.size(); ++i)
        delete(p_trigrams[i]);

    for (int i = 0; i < tri_disc_mats.size(); ++i)
        delete(tri_disc_mats[i]);

    for (int i = 0; i < tri_ungammas.size(); ++i)
        delete(tri_ungammas[i]);

    for (int i = 0; i < p_bigrams.size(); ++i)
        delete(p_bigrams[i]);

    for (int i = 0; i < bi_disc_mats.size(); ++i)
        delete(bi_disc_mats[i]);

    for (int i = 0; i < bi_ungammas.size(); ++i)
        delete(bi_ungammas[i]);
}


// This function learns a bigram PLRE
// bi_fvec is the vector of factors that determine the ranks, in general the last element of bi_fvec should be zero
// bi_pvec is the vector of powers, again the last element of bi_pvec should be zero
// bi_dvec is the vector of discounts, in general just set these all to the same value
PLRE::PLRE(SpecMoments* spec_mom,
    vector<double> bi_fvec, vector<double> bi_pvec, vector<double> bi_dvec, double uni_D)
{
    // set all the trigram stuff to null
    this->tri_fvec = new vector<double>();
    this->tri_pvec = new vector<double>();
    this->tri_dvec = new vector<double>();

    this->bi_fvec = new vector<double>(bi_fvec);
    this->bi_pvec = new vector<double>(bi_pvec);
    this->bi_dvec = new vector<double>(bi_dvec);


    this->V = spec_mom->GetV();
    this->spec_mom = spec_mom;

    this->bigram_vec = new vector<PowerBigram*>();
    this->trigram_vec = new vector<PowerTrigram*>();

    // get bigram counts, basically spec_mom->getBigrams() returns bigram probability matrix
    // need to multiply by total count (which is spec_mom->GetTotalBC() ) to get the bigram counts
    SparseMatrix<double>* b_counts = new SparseMatrix<double>(spec_mom->getBigramCMat().cast<double>());

    vector< SparseMatrix<double>* > p_bigrams; // powered matrices
    vector< SparseMatrix<double>* > bi_disc_mats; // discounts
    vector < VectorXd* > bi_ungammas; // gamma weights

    p_bigrams.push_back(b_counts); // the first bigram is the regular counts
    for (int i = 0; i < bi_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_bigram = new SparseMatrix<double>();
        SparseMatrix<double>* bi_disc_mat = new SparseMatrix<double>();
        VectorXd* bi_ungamma = new VectorXd();

        EigenHelper::Power(*b_counts, *p_bigram, bi_pvec[i]); // compute powered counts
        double prev_P = 1.0;
        if (i > 0)
            prev_P = bi_pvec[i - 1];
        // compute discounts and associated gamma weights
        DiscountComputer::ComputeDiscount(*bi_disc_mat, *bi_ungamma, *(p_bigrams[i]), *p_bigram, prev_P, bi_dvec);

        p_bigrams.push_back(p_bigram);
        bi_disc_mats.push_back(bi_disc_mat);
        bi_ungammas.push_back(bi_ungamma);
    }

    // basically append a  zero discount to the end since we don't back off to anything else
    SparseMatrix<double>* zero_bdisc = new SparseMatrix<double>(V, V);
    VectorXd* zero_gamma = new VectorXd(V);
    zero_gamma->fill(0);

    bi_disc_mats.push_back(zero_bdisc);
    bi_ungammas.push_back(zero_gamma);

    for (int i = 0; i < p_bigrams.size(); ++i)
    {
        if (i == 0)
        {
            // the first matrix is full rank
            PowerBigram* lr_bigram = new PowerFullBigram(*(p_bigrams[i]), *(bi_disc_mats[i]), *(bi_ungammas[i]));
            bigram_vec->push_back(lr_bigram);
        }
        else
        {
            cout << "computing low rank bigram\n";
            // compute low rank approximation
            PowerBigram* lr_bigram = new PowerLRBigram(*(p_bigrams[i]), *(bi_disc_mats[i]),
                *(bi_ungammas[i - 1]), *(bi_ungammas[i]), bi_fvec[i - 1]);
            cout << "done with low rank bigram\n";
            bigram_vec->push_back(lr_bigram);
        }
    }

    smoothed_unigram = new VectorXd();

    // delete intermediate quantities
    for (int i = 0; i < p_bigrams.size(); ++i)
        delete(p_bigrams[i]);

    for (int i = 0; i < bi_disc_mats.size(); ++i)
        delete(bi_disc_mats[i]);

    for (int i = 0; i < bi_ungammas.size(); ++i)
        delete(bi_ungammas[i]);
}

// computes the PLRE probability P(w = i | w_prev j)
double PLRE::GetBiCondProb(int i, int j)
{
    vector<double> vals;
    vals.reserve(3);
    vector<double> weights;
    weights.reserve(3);

    weights.push_back(1);
    for (int n = 0; n < bigram_vec->size(); ++n)
    {
        double val = bigram_vec->at(n)->CondLRDiscAt(i, j);
        vals.push_back(val);

        if (n < bigram_vec->size() - 1)
        {
            double curr_w = weights[n];
            double new_w = curr_w * bigram_vec->at(n)->GammaAt(j);
            weights.push_back(new_w);
        }
    }



    assert(vals.size() == weights.size());

    double total_val = 0;

    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }

    assert(total_val >= 0 && total_val <= (1 + 1e-10));

    return total_val;
}




double PLRE::GetTriBiCondProb(int i, int j, int k)
{
    vector<double> vals;
    vector<double> weights;

    weights.push_back(1);
    for (int n = 0; n < trigram_vec->size(); ++n)
    {
        double val = trigram_vec->at(n)->CondLRDiscAt(i, j, k);
        vals.push_back(val);

        double curr_w = weights[n];
        double new_w = curr_w * trigram_vec->at(n)->GammaAt(j, k);
        weights.push_back(new_w);
    }

    double val = bigram_vec->at(0)->CondLRDiscAt(i, j);
    vals.push_back(val);


    assert(vals.size() == weights.size());

    double total_val = 0;

    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }

    assert(total_val >= 0 && total_val <= (1 + 1e-10));

    return total_val;
}

double PLRE::GetCondProb(int i, int j, int k)
{
    vector<double> vals;
    vals.reserve(5);
    vector<double> weights;
    weights.reserve(5);
    weights.push_back(1);
    for (int n = 0; n < trigram_vec->size(); ++n)
    {
        double val = trigram_vec->at(n)->CondLRDiscAt(i, j, k);
        vals.push_back(val);

        double curr_w = weights[n];
        double new_w = curr_w * trigram_vec->at(n)->GammaAt(j, k);
        weights.push_back(new_w);
    }

    for (int n = 0; n < bigram_vec->size(); ++n)
    {
        if (n < bigram_vec->size() - 1)
        {
            double val = bigram_vec->at(n)->CondLRDiscAt(i, j);
            vals.push_back(val);

            double curr_w = weights[weights.size() - 1];
            double new_w = curr_w * bigram_vec->at(n)->GammaAt(j);
            weights.push_back(new_w);
        }
        else
        {
            double val = smoothed_unigram->coeff(i);
            vals.push_back(val);
        }
    }

    assert(vals.size() == weights.size());

    double total_val = 0;
    string val_str = "";
    string w_str = "";
    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }

    assert(total_val >= 0 && total_val <= (1 + 1e-10));

    return total_val;
}

void PLRE::ComputeBCounts(SparseMatrix<double>& b_counts, SMat3Tensor<int>& trigram_counts,
    SparseMatrix<double>& zero_trigram)
{
    b_counts.resize(V, V);

    map<pair<int, int>, double> total_counts;

    int mode = trigram_counts.GetMode();
    assert(mode == 0);

    for (int i = 0; i < zero_trigram.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(zero_trigram, i); it; ++it)
        {
            int i1 = it.row();
            int i2 = it.col();
            double val = it.value();
            vector<int> indices = trigram_counts.GetTensorIndex(i1, i2);

            pair<int, int> p(indices[0], indices[1]);

            if (total_counts.count(p) == 0)
                total_counts[p] = 0;

            total_counts[p] += val;

        }
    }

    map<pair<int, int>, double>::iterator iter;
    vector<triplet> triplets;
    for (iter = total_counts.begin(); iter != total_counts.end(); ++iter)
    {
        int r = iter->first.first;
        int c = iter->first.second;
        double val = (iter->second);
        triplets.push_back(triplet(r, c, val));
    }

    b_counts.setFromTriplets(triplets.begin(), triplets.end());
}


void PLRE::ComputeUCounts(VectorXd& u_counts, SparseMatrix<double>& zero_counts)
{
    u_counts.resize(V);
    u_counts.fill(0);
    map<int, double> total_counts;

    for (int i = 0; i < zero_counts.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(zero_counts, i); it; ++it)
        {
            int i1 = it.row();
            int i2 = it.col();
            double val = it.value();

            u_counts(i1) += val;

        }
    }
}

void PLRE::ComputeSmoothedUnigram(VectorXd& u_counts, double uni_D)
{
    int non_zero = 0;
    for (int i = 0; i < u_counts.size(); ++i)
    {
        if (u_counts(i) > 0)
            non_zero += 1;
    }

    double sum = u_counts.sum();
    double gamma = non_zero * uni_D / sum;

    assert(gamma <= 1);

    smoothed_unigram = new VectorXd(V);
    smoothed_unigram->fill(0);

    for (int i = 0; i < u_counts.size(); ++i)
    {
        double val = u_counts.coeff(i);

        if (val > 0)
        {
            assert(val >= 1);
            (*smoothed_unigram)(i) = (val - uni_D) / sum + gamma / V;
        }
        else
        {
            (*smoothed_unigram)(i) = gamma / V;
        }
    }
    double smooth_sum = smoothed_unigram->sum();
    assert(abs(smooth_sum - 1.0) < 1e-8);
}
