/* Ankur Parikh
  PLRE.hpp

  learns 3rd order and 2nd order PLRE models */


#pragma once 

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"

#include <limits>
#include <vector>
#include <set>
#include "PowerLRBigram.hpp"
#include "PowerLRTrigram.hpp"
#include "PowerFullBigram.hpp"
#include "PowerFullTrigram.hpp"
#include "PowerBigram.hpp"
#include "PowerTrigram.hpp"
#include "PowerQuadgram.hpp"
#include "eigen_boost_serialization.hpp"

using namespace std;
using namespace Eigen;


class PLRE
{

private:

    SpecMoments* spec_mom;
    VectorXd* smoothed_unigram;
    vector<PowerBigram*>* bigram_vec;
    vector<PowerTrigram*>* trigram_vec;

    // f controls rank e.g. V/f = rank
    vector<double>* tri_fvec;
    vector<double>* bi_fvec;

    // vector of powers 
    vector<double>* tri_pvec;
    vector<double>* bi_pvec;

    // vector of discounts
    vector<double>* tri_dvec;
    vector<double>* bi_dvec;

    // vocabulary size
    int V;

    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & spec_mom;
        ar & smoothed_unigram;

        ar & tri_fvec;
        ar & bi_fvec;

        ar & tri_pvec;
        ar & bi_pvec;

        ar & tri_dvec;
        ar & bi_dvec;

        ar & V;

        int bi_size = bigram_vec->size();
        ar & bi_size;

        if (bi_size > 0)
        {
            PowerFullBigram* pfb = (PowerFullBigram*)(bigram_vec->at(0));
            ar & pfb;
            for (int i = 1; i < bigram_vec->size(); ++i)
            {
                PowerLRBigram* plrb = (PowerLRBigram*)(bigram_vec->at(i));
                ar & plrb;
            }
        }

        int tri_size = trigram_vec->size();
        ar & tri_size;
        if (tri_size > 0)
        {
            PowerFullTrigram* pft = (PowerFullTrigram*)(trigram_vec->at(0));
            ar & pft;
            for (int i = 1; i < trigram_vec->size(); ++i)
            {
                PowerLRTrigram* plrt = (PowerLRTrigram*)(trigram_vec->at(i));
                ar & plrt;
            }
        }
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        ar & spec_mom;
        ar & smoothed_unigram;

        ar & tri_fvec;
        ar & bi_fvec;

        ar & tri_pvec;
        ar & bi_pvec;

        ar & tri_dvec;
        ar & bi_dvec;

        ar & V;

        bigram_vec = new vector<PowerBigram*>();
        trigram_vec = new vector<PowerTrigram*>();

        int bi_size;
        ar & bi_size;

        for (int i = 0; i < bi_size; ++i)
        {
            if (i == 0)
            {
                PowerFullBigram* pfb;
                ar & pfb;
                bigram_vec->push_back(pfb);
            }
            else
            {
                PowerLRBigram* plrb;
                ar & plrb;
                bigram_vec->push_back(plrb);
            }
        }

        int tri_size;
        ar & tri_size;

        for (int i = 0; i < tri_size; ++i)
        {
            if (i == 0)
            {
                PowerFullTrigram* pft;
                ar & pft;
                trigram_vec->push_back(pft);
            }
            else
            {
                PowerLRTrigram* plrt;
                ar & plrt;
                trigram_vec->push_back(plrt);
            }
        }

        if (tri_size > 1)
        {
            delete(bigram_vec->at(bigram_vec->size() - 1));
            bigram_vec->at(bigram_vec->size() - 1) = new PowerLRBigram();
        }
    }



    BOOST_SERIALIZATION_SPLIT_MEMBER()

public:

    PLRE()
    {
        spec_mom = NULL;

        tri_fvec = NULL;
        bi_fvec = NULL;

        tri_pvec = NULL;
        bi_pvec = NULL;

        tri_dvec = NULL;
        bi_dvec = NULL;

        bigram_vec = NULL;
        trigram_vec = NULL;
    }


    PLRE(SpecMoments* spec_mom, vector<double> tri_fvec, vector<double> bi_fvec,
        vector<double> tri_pvec, vector<double> bi_pvec,
        vector<double> tri_dvec, vector<double> bi_dvec, double uni_D);

    PLRE(SpecMoments* spec_mom, vector<double> bi_fvec, vector<double> bi_pvec,
        vector<double> bi_dvec, double uni_D);


    ~PLRE()
    {
        delete(smoothed_unigram);
        delete(tri_fvec);
        delete(bi_fvec);
        delete(tri_pvec);
        delete(bi_pvec);
        delete(tri_dvec);
        delete(bi_dvec);

        for (int i = 0; i < bigram_vec->size(); ++i)
            delete(bigram_vec->at(i));
        for (int i = 0; i < trigram_vec->size(); ++i)
            delete(trigram_vec->at(i));

        delete(bigram_vec);
        delete(trigram_vec);
    }

    PowerLRBigram& GetPowerLRBigram()
    {
        assert(bigram_vec->size() == 3);
        PowerLRBigram* plrb = (PowerLRBigram*)(bigram_vec->at(1));
        return *plrb;
    }

    double GetCondProb(int i, int j, int k);

    double GetBiCondProb(int i, int j);

    double GetTriBiCondProb(int i, int j, int k);

    double GetCondProbPrint(int i, int j, int k);

    void ComputeBCounts(SparseMatrix<double>& b_counts, SMat3Tensor<int>& trigram_counts,
        SparseMatrix<double>& zero_trigram);

    void ComputeUCounts(VectorXd& u_counts, SparseMatrix<double>& bigram_counts);

    void ComputeSmoothedUnigram(VectorXd& u_counts, double uni_D);

};
