/* Ankur Parikh
   PLREWrapper.hpp

   wrapper around PLRE model for query probabilities efficiently,
   has a cache so that repeated queries can simply be looked up in the cache
   */


#pragma once

#include "QuadPLRE.hpp"
#include "PLRE.hpp"

#include "eigen_boost_serialization.hpp"
#include <boost/tr1/unordered_map.hpp>

using namespace std;
using namespace Eigen;


class PLREWrapper
{

private:

    SpecMoments* spec_mom; // moments
    PLRE* bi_plre; // bigram power low rank ensemble
    PLRE* tri_plre; // trigram power low rank ensemble
    QuadPLRE* quad_plre; // quadgraph power low rank ensemble

    tr1::unordered_map<iquadple, double, hash_Q>* quad_cache; // cache for quadgram queries (i,j,k,l)
    tr1::unordered_map<itriple, double, hash_T>* triple_cache; //cache for trigram queries (i,j,k)
    tr1::unordered_map<pair<int, int>, double, hash_P>* pair_cache; // cache for bigram queries (i,j)

    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & spec_mom;
        ar & bi_plre;
        ar & tri_plre;
        ar & quad_plre;
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        bi_plre = NULL;
        tri_plre = NULL;
        quad_plre = NULL;
        ar & spec_mom;
        ar & bi_plre;
        ar & tri_plre;
        ar & quad_plre;

        quad_cache = new  tr1::unordered_map<iquadple, double, hash_Q>();
        triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
        pair_cache = new tr1::unordered_map<pair<int, int>, double, hash_P>();
    }



    BOOST_SERIALIZATION_SPLIT_MEMBER()


public:

    PLREWrapper()
    {
        bi_plre = NULL;
        tri_plre = NULL;
        quad_plre = NULL;

        /*	quad_cache = new  tr1::unordered_map<iquadple,double, hash_Q>();
            triple_cache = new tr1::unordered_map<itriple, double, hash_T>();
            pair_cache = new tr1::unordered_map<pair<int,int>, double, hash_P>();*/
    }
    PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre, PLRE* tri_plre, QuadPLRE* quad_plre);
    PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre, PLRE* tri_plre);
    PLREWrapper(SpecMoments* spec_mom, PLRE* bi_plre);

    ~PLREWrapper()
    {
        if (bi_plre != NULL)
            delete(bi_plre);
        if (tri_plre != NULL)
            delete(tri_plre);
        if (quad_plre != NULL)
            delete(quad_plre);
        delete(quad_cache);
        delete(triple_cache);
        delete(pair_cache);
    }

    // queries with strings
    double GetCondProb(const string& w, const string& wp, const string& wpp, const string& wppp);
    double GetCondProb(const string& w, const string& wp, const string& wpp);
    double GetCondProb(const string& w, const string& wp);
    double GetCondProb(const string& w)
    {
        int i = spec_mom->GetFinalID(w);
        return GetCondProb(i);
    }

    // queries with word indices
    double GetCondProb(int i, int j, int k, int l);
    double GetCondProb(int i, int j, int k);
    double GetCondProb(int i, int j);
    double GetCondProb(int i)
    {
        VectorXd& unigrams = spec_mom->getUnigrams();
        double prob = unigrams(i);
        return prob;
    }

    // test if string is out of vocabulary
    bool IsOOV(const string& w) { return (spec_mom->GetFinalID(w) == spec_mom->GetOOVIndex()); }
};