/* Ankur Parikh
   PowerLRBigram.cpp

   see .hpp for class/function comments
   */


#include "PowerLRBigram.hpp"
#include "EigenHelper.hpp"
#include "EM.hpp"
#include "CondensedSMat.hpp"
#include <algorithm> 

using namespace std;

bool pairCompare(const pair<int, double>& firstElem, const pair<int, double>& secondElem)
{
    return (firstElem.second < secondElem.second);
}

PowerLRBigram::PowerLRBigram()
{
    l_cpt = new MatrixXd();
    r_cpt = new MatrixXd();
    real_l_cpt = new MatrixXd();
    denom = new VectorXd();
    after_gamma = new VectorXd();
}

PowerLRBigram::~PowerLRBigram()
{
    delete(denom);
    delete(l_cpt);
    delete(r_cpt);
    delete(after_gamma);
    delete(real_l_cpt);
}

PowerLRBigram::PowerLRBigram(SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
    VectorXd& before_gamma, VectorXd& after_gamma, double f)
{
    Initialize(power_counts, disc, before_gamma, after_gamma, f);
}

void PowerLRBigram::Initialize(SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
    VectorXd& before_gamma, VectorXd& after_gamma, double f)
{
    l_cpt = new MatrixXd();
    r_cpt = new MatrixXd();
    real_l_cpt = new MatrixXd();
    denom = new VectorXd();
    this->f = f;
    this->after_gamma = new VectorXd(after_gamma);

    EigenHelper::ColSums(power_counts, *denom);

    SparseMatrix<double> disc_counts = power_counts - disc;

    VectorXd denom_inv = denom->array().max(1e-8).inverse().matrix();

    int full_r = min(disc_counts.rows(), disc_counts.cols());
    int r = (int)(f * full_r);
    r = max(r, 1);

    EM::RunBigParallelEM(*l_cpt, *r_cpt, *real_l_cpt, disc_counts, r, EM::MAX_ITERS, EM::NUM_CORES);
    VectorXd gamma_inv = before_gamma.array().max(1e-8).inverse().matrix();
}


double PowerLRBigram::LRDiscAt(int i, int j)
{
    double val = l_cpt->row(i) * r_cpt->col(j);
    return val;
}

double PowerLRBigram::CondLRDiscAt(int i, int j)
{
    pair<int, int> p(i, j);
    double d = (*denom)(j);
    if (d == 0.0)
        return 0;
    else
    {
        double n = LRDiscAt(i, j);
        double val = n / d;
        return val;
    }
}

double PowerLRBigram::GammaAt(int j)
{
    double d = (*denom)(j);
    if (d == 0.0)
    {
        return 1;
    }
    else
        return (*after_gamma)(j) / d;
}

vector<pair<int, double> > PowerLRBigram::TopLeftIndices(int index, int num_top)
{
    vector< pair<int, double> > vals;
    vals.reserve(real_l_cpt->rows());

    VectorXd index_vec = real_l_cpt->row(index);


    for (int i = 0; i < real_l_cpt->rows(); ++i)
    {
        VectorXd i_vec = real_l_cpt->row(i);
        double dot = (i_vec.transpose() * index_vec);
        dot /= ((index_vec.norm() * i_vec.norm()));
        vals.push_back(pair<int, double>(i, -1 * dot));
    }


    sort(vals.begin(), vals.end(), pairCompare);

    vector< pair<int, double> > top_indices;
    for (int i = 0; i < num_top; ++i)
        top_indices.push_back(vals[i]);

    return top_indices;
}


vector<pair<int, double> > PowerLRBigram::TopRightIndices(int index, int num_top)
{
    vector< pair<int, double> > vals;
    vals.reserve(r_cpt->cols());

    VectorXd index_vec = r_cpt->col(index);


    for (int i = 0; i < r_cpt->cols(); ++i)
    {
        VectorXd i_vec = r_cpt->col(i);
        double dot = (i_vec.transpose() * index_vec);
        dot /= (index_vec.norm() * i_vec.norm());
        vals.push_back(pair<int, double>(i, -1 * dot));
    }


    sort(vals.begin(), vals.end(), pairCompare);

    vector< pair<int, double> > top_indices;
    for (int i = 0; i < num_top; ++i)
        top_indices.push_back(vals[i]);

    return top_indices;
}
