/* Ankur Parikh
   PowerLRQuadgram.cpp

   see class/function comments in .hpp
   */

#include "PowerLRQuadgram.hpp"
#include "EM.hpp"
#include "EigenHelper.hpp"
#include "CondensedSMat.hpp"
#include "CondensedDMat.hpp"
#include "IOLibrary.hpp"

// constructor / initializer
PowerLRQuadgram::PowerLRQuadgram()
{
    l_cpts = NULL;
    r_cpts = NULL;

    count_array = NULL;

    denom_ten = NULL;
    after_gamma_ten = NULL;

    counts = NULL;
    f = -1;
}

PowerLRQuadgram::~PowerLRQuadgram()
{
    delete(denom_ten);
    delete(after_gamma_ten);

    map<pair<int, int>, CondensedDMat*>::iterator iter;

    for (iter = l_cpts->begin(); iter != l_cpts->end(); ++iter)
        delete(iter->second);

    for (iter = r_cpts->begin(); iter != r_cpts->end(); ++iter)
        delete(iter->second);

    delete(l_cpts);
    delete(r_cpts);
}

PowerLRQuadgram::PowerLRQuadgram(SMat4Tensor<int>* counts, SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
    VectorXd& before_gamma, VectorXd& after_gamma, double f)
{
    Initialize(counts, power_counts, disc, before_gamma, after_gamma, f);
}

void PowerLRQuadgram::Initialize(SMat4Tensor<int>* counts, SparseMatrix<double>& power_counts,
    SparseMatrix<double>& disc, VectorXd& before_gamma, VectorXd& after_gamma, double f)
{
    this->counts = counts;

    denom_ten = new VectorXd();
    EigenHelper::ColSums(power_counts, *denom_ten);
    this->after_gamma_ten = new VectorXd(after_gamma);
    SparseMatrix<double> disc_counts = power_counts - disc;

    // need to rearrange the matrix for fast accessing
    vector<int> dims = counts->Dims();
    int V = dims[0];
    count_array = new SQuadMatrixArray<double>();
    count_array->Initialize(dims, 1, 2);
    count_array->startFill();

    this->l_cpts = new map<pair<int, int>, CondensedDMat*>();
    this->r_cpts = new map<pair<int, int>, CondensedDMat*>();

    vector< vector<int>* > mid_indices(V, NULL);
    set < pair<int, int> > mid_set;
    set<int> covered_set;
    map<itriple, int>& triple_2_index = counts->GetTriple2IndexMap();
    map<itriple, int>::iterator iter1;
    for (iter1 = triple_2_index.begin(); iter1 != triple_2_index.end(); ++iter1)
    {
        itriple t = iter1->first;
        int c = iter1->second;
        Eigen::SparseVector<double> vec = disc_counts.col(c);
        for (SparseVector<double>::InnerIterator it(vec); it; ++it)
        {
            int r = it.row();
            double val = it.value();
            if (val < -1e-10)
                assert(0);
            if (val < 0)
                val = 0;

            int ind1 = r;
            int ind2 = t.get<0>();
            int ind3 = t.get<1>();
            int ind4 = t.get<2>();

            pair<int, int> midp(ind2, ind3);
            if (mid_set.count(midp) == 0)
            {
                if (mid_indices.at(ind2) == NULL)
                {
                    mid_indices[ind2] = new vector<int>();
                }
                mid_indices[ind2]->push_back(ind3);
                mid_set.insert(midp);
            }

            double d = denom_ten->coeff(c);
            if (denom_ten->coeff(c) == 0.0)
                assert(val == 0.0);

            count_array->Set(ind1, ind2, ind3, ind4, val);
        }
    }
    count_array->endFill();

    // compute all the low rank approximations for each slice
#pragma omp parallel for
    for (int i = 0; i < mid_indices.size(); ++i)
    {
        if (mid_indices[i] == NULL)
            continue;

        vector<int>& other_indices = *(mid_indices[i]);


        SparseMatrix<double, RowMajor> d_mat;
        SparseMatrix<double, RowMajor> g_mat;


        for (int j = 0; j < other_indices.size(); ++j)
        {
            int ind2 = i;
            int ind3 = other_indices.at(j);

            pair<int, int> midp(ind2, ind3);

            CondensedSMat csmat;
#pragma omp critical
            {
                count_array->GetMat(csmat, 1, 2, ind2, ind3);
            }

            MatrixXd l_cpt;
            MatrixXd r_cpt;

            if (csmat.Sum() == 0) // the whole slice is zero
            {
                map<int, int> row_map;
                map<int, int> col_map;

                CondensedDMat* c_lcpt = new CondensedDMat(l_cpt, row_map, 0);
                CondensedDMat* c_rcpt = new CondensedDMat(r_cpt, col_map, 1);

#pragma omp critical
                {
                    (*l_cpts)[midp] = c_lcpt;
                    (*r_cpts)[midp] = c_rcpt;
                }

            }
            else // slice is non zero so take low rank approximation
            {
                SparseMatrix<double>& c_m = csmat.GetCondMat();

                map<int, int>& row_map = csmat.GetRowMap();
                map<int, int>& col_map = csmat.GetColMap();

                int full_r = min(c_m.rows(), c_m.cols());
                int r = (int)(f * full_r);
                r = max(r, 1);
                EM::RunBigEM(l_cpt, r_cpt, c_m, r, EM::MAX_ITERS / 2);

                CondensedDMat* c_lcpt = new CondensedDMat(l_cpt, row_map, 0);
                CondensedDMat* c_rcpt = new CondensedDMat(r_cpt, col_map, 1);


#pragma omp critical
                {
                    (*l_cpts)[midp] = c_lcpt;
                    (*r_cpts)[midp] = c_rcpt;
                }
            }
        }
    }

    for (int i = 0; i < mid_indices.size(); ++i)
    {
        if (mid_indices[i] != NULL)
            delete(mid_indices[i]);
    }

    delete(count_array);
    count_array = NULL;
}

double PowerLRQuadgram::LRDiscAt(int i, int j, int k, int l)
{
    pair<int, int> midp(j, k);
    if (l_cpts->count(midp) == 0 || r_cpts->count(midp) == 0.0)
        return 0;

    CondensedDMat& l_cpt = *((*l_cpts)[midp]);
    CondensedDMat& r_cpt = *((*r_cpts)[midp]);

    if (!l_cpt.ContainsRow(i) || !r_cpt.ContainsCol(l))
        return 0.0;


    double val = l_cpt.Row(i).transpose() * r_cpt.Col(l);

    if (val != val)
    {
        cout << "wtf: " << val << "\n";
        cout << l_cpt.Row(i) << "\n";
        cout << r_cpt.Col(l) << "\n";
    }

    return val;
}

double PowerLRQuadgram::CondLRDiscAt(int i, int j, int k, int l)
{
    double n = LRDiscAt(i, j, k, l);

    //	MatrixXd& l_cpt = *(l_cpts->at(j));
    //	MatrixXd& r_cpt = *(r_cpts->at(j));

    //	VectorXd v = l_cpt * r_cpt.col(k);
    //	double d = v.sum();
    int index = counts->ComputeAltIndex(j, k, l);
    if (index == -1)
        return 0.0;
    double d = denom_ten->coeff(index);
    if (d == 0.0)
        return 0.0;
    double val = n / d;

    if (val > 1.0 || (n != n) || (d != d))
    {
        assert(0);
    }
    return val;
}

double PowerLRQuadgram::GammaAt(int j, int k, int l)
{
    int index = counts->ComputeAltIndex(j, k, l);
    if (index == -1)
        return 1.0;
    double g = after_gamma_ten->coeff(index);
    double d = denom_ten->coeff(index);

    if (d == 0.0)
        return 1.0;
    else
        return g / d;
}