/* Ankur Parikh
   PowerLRTrigram.cpp

   see .hpp for class/function comments
   */

#include "PowerLRTrigram.hpp"
#include "EM.hpp"
#include "EigenHelper.hpp"
#include "CondensedSMat.hpp"
#include "CondensedDMat.hpp"
#include "IOLibrary.hpp"
#include <iostream>    
#include <algorithm>         


void PrintMap(map<int, int>& m)
{
    map<int, int>::iterator iter;
    for (iter = m.begin(); iter != m.end(); ++iter)
    {
        cout << iter->first << ": " << iter->second << "\t";
    }
    cout << "\n";
}

// constructor / initializer
PowerLRTrigram::PowerLRTrigram()
{
    l_cpts = NULL;
    r_cpts = NULL;

    count_array = NULL;

    denoms = NULL;
    after_gammas = NULL;

    counts = NULL;
    K = -1;
}

PowerLRTrigram::~PowerLRTrigram()
{
    delete(denoms);
    delete(after_gammas);
    //	delete(cache);
    for (int i = 0; i < l_cpts->size(); ++i)
    {
        if (l_cpts->at(i) != NULL)
            delete(l_cpts->at(i));
    }
    for (int i = 0; i < r_cpts->size(); ++i)
    {
        if (r_cpts->at(i) != NULL)
            delete(r_cpts->at(i));
    }
    delete(l_cpts);
    delete(r_cpts);
}

PowerLRTrigram::PowerLRTrigram(SMat3Tensor<int>* counts, SparseMatrix<double>& power_counts, SparseMatrix<double>& disc,
    VectorXd& before_gamma, VectorXd& after_gamma, int K, double f)
{
    Initialize(counts, power_counts, disc, before_gamma, after_gamma, K, f);
}

void PowerLRTrigram::Initialize(SMat3Tensor<int>* counts, SparseMatrix<double>& power_counts,
    SparseMatrix<double>& disc, VectorXd& before_gamma, VectorXd& after_gamma, int K, double f)
{


    this->counts = counts;

    denoms = new VectorXd();
    EigenHelper::ColSums(power_counts, *denoms);
    SparseMatrix<double> disc_counts = power_counts - disc;

    // need to rearrage the count tensor to make it more amenable to slicing on w_{i-1}

    vector<int> dims = counts->Dims();
    count_array = new SMatrixArray<double>();
    count_array->Initialize(dims, 1);
    count_array->startFill();

    this->l_cpts = new vector<CondensedDMat*>(dims[1], NULL);
    this->r_cpts = new vector<CondensedDMat*>(dims[1], NULL);

    after_gammas = new VectorXd(after_gamma);

    vector<triplet> g1_triplets;
    vector<triplet> g2_triplets;
    vector<triplet> d_triplets;

    set<int> covered_set;
    for (int i = 0; i < disc_counts.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(disc_counts, i); it; ++it)
        {
            int i1 = it.row();
            int i2 = it.col();
            double val = it.value();

            if (val < -1e-10)
                assert(0);

            if (val < 0)
                val = 0;

            vector<int> indices = counts->GetTensorIndex(i1, i2);

            if (denoms->coeff(i2) == 0.0)
                assert(val == 0.0);

            count_array->Set(indices[0], indices[1], indices[2], val);
        }
    }
    count_array->endFill();

    // randomly shuffle indices to easily get parallelization
    vector<int> temp_indices;
    VectorPlus::Seq(temp_indices, 0, 1, dims[1]);

    vector<int> high_rank_indices;

    random_shuffle(temp_indices.begin(), temp_indices.end());

    // get low rank approximations
#pragma omp parallel for num_threads(8)
    for (int j = 0; j < dims[1]; ++j)
    {
        int index = temp_indices[j];
        CondensedSMat csmat;
#pragma omp critical
        {
            // get slice
            count_array->GetMat(csmat, 1, index);
        }

        MatrixXd l_cpt;
        MatrixXd r_cpt;

        if (csmat.Sum() == 0) // slice is all zero
        {
            map<int, int> row_map;
            map<int, int> col_map;

            CondensedDMat* c_lcpt = new CondensedDMat(l_cpt, row_map, 0);
            CondensedDMat* c_rcpt = new CondensedDMat(r_cpt, col_map, 1);

#pragma omp critical
            {
                l_cpts->at(index) = c_lcpt;
                r_cpts->at(index) = c_rcpt;
            }
        }
        else
        {
            SparseMatrix<double>& c_m = csmat.GetCondMat();

            map<int, int>& row_map = csmat.GetRowMap();
            map<int, int>& col_map = csmat.GetColMap();

            int full_r = min(c_m.rows(), c_m.cols());
            int r = (int)(f * full_r);
            r = max(r, 1);
            if (r >= 80) // needs to be processed later since has high rank
            {
#pragma omp critical
                {
                    high_rank_indices.push_back(index);
                }
            }
            else // has low rank so can be processed in parallel
            {
                EM::RunBigEM(l_cpt, r_cpt, c_m, r, EM::MAX_ITERS / 2);

                CondensedDMat* c_lcpt = new CondensedDMat(l_cpt, row_map, 0);
                CondensedDMat* c_rcpt = new CondensedDMat(r_cpt, col_map, 1);

#pragma omp critical
                {
                    l_cpts->at(index) = c_lcpt;
                    r_cpts->at(index) = c_rcpt;
                }
            }
        }
    }

    // process the high rank slices sequentially
    for (int j = 0; j < high_rank_indices.size(); ++j)
    {
        int index = high_rank_indices[j];
        CondensedSMat csmat;
#pragma omp critical
        {
            count_array->GetMat(csmat, 1, index);
        }


        MatrixXd l_cpt;
        MatrixXd r_cpt;

        if (csmat.Sum() == 0)
        {
            assert(0);
        }
        else
        {
            SparseMatrix<double>& c_m = csmat.GetCondMat();

            map<int, int>& row_map = csmat.GetRowMap();
            map<int, int>& col_map = csmat.GetColMap();

            int full_r = min(c_m.rows(), c_m.cols());
            int r = (int)(f * full_r);
            r = max(r, 1);
            r = min(r, 300);
            if (r > 1)
            {
                EM::RunBigParallelEM(l_cpt, r_cpt, c_m, r, EM::MAX_ITERS / 2, EM::NUM_CORES);


                CondensedDMat* c_lcpt = new CondensedDMat(l_cpt, row_map, 0);
                CondensedDMat* c_rcpt = new CondensedDMat(r_cpt, col_map, 1);

#pragma omp critical
                {
                    l_cpts->at(index) = c_lcpt;
                    r_cpts->at(index) = c_rcpt;
                }
            }
            else
            {
                assert(0);
            }
        }
    }

    delete(count_array);
}

double PowerLRTrigram::LRDiscAt(int i, int j, int k)
{
    CondensedDMat& l_cpt = *(l_cpts->at(j));
    CondensedDMat& r_cpt = *(r_cpts->at(j));

    if (!l_cpt.ContainsRow(i) || !r_cpt.ContainsCol(k))
        return 0.0;

    double val = l_cpt.Row(i).transpose() * r_cpt.Col(k);

    // some NaN issue
    if (val != val)
    {
        cout << "wtf: " << val << "\n";
        cout << l_cpt.Row(i) << "\n";
        cout << r_cpt.Col(k) << "\n";
        assert(0);

    }
    return val;
}

double PowerLRTrigram::CondLRDiscAt(int i, int j, int k)
{
    int index = counts->ComputeAltIndex(j, k);
    if (index == -1)
        return 0.0;

    double d = denoms->coeff(index);
    if (d == 0.0)
        return 0.0;

    double n = LRDiscAt(i, j, k);
    double val = n / d;

    return val;
}

double PowerLRTrigram::GammaAt(int j, int k)
{
    int index = counts->ComputeAltIndex(j, k);
    if (index == -1)
        return 1.0;

    double g = after_gammas->coeff(index);
    double d = denoms->coeff(index);

    if (d == 0.0)
        return 1.0;
    else
        return g / d;
}