/* Ankur Parikh
  PLRE.cpp

  learns 4th order PLRE model */

#include "QuadPLRE.hpp"
#include "IOLibrary.hpp"
#include "DiscountComputer.hpp"
#include "EM.hpp"
#include "PowerLRQuadgram.hpp"
#include "PowerFullQuadgram.hpp"
#include "PowerFullTrigram.hpp"
#include "PowerFullBigram.hpp"
#include "EigenHelper.hpp"

using namespace std;
using namespace Eigen;



QuadPLRE::QuadPLRE(SpecMoments* spec_mom,
    vector<double> quad_fvec, vector<double> tri_fvec, vector<double> bi_fvec,
    vector<double> quad_pvec, vector<double> tri_pvec, vector<double> bi_pvec,
    vector<double> quad_dvec, vector<double> tri_dvec, vector<double> bi_dvec, double uni_D)
{
    this->quad_fvec = new vector<double>(quad_fvec);
    this->tri_fvec = new vector<double>(tri_fvec);
    this->bi_fvec = new vector<double>(bi_fvec);

    this->quad_pvec = new vector<double>(quad_pvec);
    this->tri_pvec = new vector<double>(tri_pvec);
    this->bi_pvec = new vector<double>(bi_pvec);

    this->quad_dvec = new vector<double>(quad_dvec);
    this->tri_dvec = new vector<double>(tri_dvec);
    this->bi_dvec = new vector<double>(bi_dvec);

    this->V = spec_mom->GetV();
    this->spec_mom = spec_mom;

    this->bigram_vec = new vector<PowerBigram*>();
    this->trigram_vec = new vector<PowerTrigram*>();
    this->quadgram_vec = new vector<PowerQuadgram*>();

    SMat4Tensor<int>& quadgram_counts = spec_mom->getQuadgramCMat();
    SparseMatrix<int>& qc_counts = quadgram_counts.GetMat();
    SparseMatrix<double> qc_double = qc_counts.cast<double>();
    SparseMatrix<double>* quadgram_cmat = new SparseMatrix<double>(qc_double);

    vector< SparseMatrix<double>* > p_quadgrams;
    vector< SparseMatrix<double>* > quad_disc_mats;
    vector < VectorXd* > quad_ungammas;

    // get powered/discounted quadgrams
    p_quadgrams.push_back(quadgram_cmat);
    for (int i = 0; i < quad_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_quadgram = new SparseMatrix<double>();
        SparseMatrix<double>* quad_disc_mat = new SparseMatrix<double>();
        VectorXd* quad_ungamma = new VectorXd();

        EigenHelper::Power(*quadgram_cmat, *p_quadgram, quad_pvec[i]);

        double prev_P = 1.0;
        if (i > 0)
            prev_P = quad_pvec[i - 1];

        DiscountComputer::ComputeDiscount(*quad_disc_mat, *quad_ungamma, *(p_quadgrams[i]), *p_quadgram,
            prev_P, quad_dvec);

        p_quadgrams.push_back(p_quadgram);
        quad_disc_mats.push_back(quad_disc_mat);
        quad_ungammas.push_back(quad_ungamma);
    }
    assert(quad_pvec[quad_pvec.size() - 1] == 0);


    // compute low rank quadgrams
    VectorXd q_gamma;
    for (int i = 0; i < p_quadgrams.size() - 1; ++i)
    {
        if (i == 0)
        {
            PowerQuadgram* lr_quadgram = new PowerFullQuadgram(&quadgram_counts, *(quad_disc_mats[i]), *(quad_ungammas[i]));
            quadgram_vec->push_back(lr_quadgram);
            q_gamma = *(quad_ungammas[i]);
        }
        else
        {
            cout << "computing low rank quadgram\n";

            PowerLRQuadgram* lr_quadgram = new PowerLRQuadgram(&quadgram_counts, *(p_quadgrams[i]), *(quad_disc_mats[i]),
                q_gamma, *(quad_ungammas[i]), quad_fvec[i - 1]);
            cout << "done computing low rank quadgram\n";
            quadgram_vec->push_back(lr_quadgram);

            q_gamma = *(quad_ungammas[i]);
        }

    }

    // trigram is power 0, rank 1 quadgram
    t_counts = new SMat3Tensor<int>();
    ComputeTCounts(*t_counts, quadgram_counts, *(p_quadgrams[p_quadgrams.size() - 1]));

    for (int i = 0; i < p_quadgrams.size(); ++i)
        delete(p_quadgrams[i]);

    for (int i = 0; i < quad_disc_mats.size(); ++i)
        delete(quad_disc_mats[i]);

    for (int i = 0; i < quad_ungammas.size(); ++i)
        delete(quad_ungammas[i]);

    SparseMatrix<double>* trigram_cmat = new SparseMatrix<double>(t_counts->GetMat().cast<double>());

    vector< SparseMatrix<double>* > p_trigrams;
    vector< SparseMatrix<double>* > tri_disc_mats;
    vector < VectorXd* > tri_ungammas;

    // get powered /discounted trigrams
    p_trigrams.push_back(trigram_cmat);
    for (int i = 0; i < tri_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_trigram = new SparseMatrix<double>();
        SparseMatrix<double>* tri_disc_mat = new SparseMatrix<double>();
        VectorXd* tri_ungamma = new VectorXd();

        EigenHelper::Power(*trigram_cmat, *p_trigram, tri_pvec[i]);

        double prev_P = 1.0;
        if (i > 0)
            prev_P = tri_pvec[i - 1];

        DiscountComputer::ComputeDiscount(*tri_disc_mat, *tri_ungamma, *(p_trigrams[i]), *p_trigram,
            prev_P, tri_dvec);

        p_trigrams.push_back(p_trigram);
        tri_disc_mats.push_back(tri_disc_mat);
        tri_ungammas.push_back(tri_ungamma);
    }
    assert(tri_pvec[tri_pvec.size() - 1] == 0);

    // low rank trigrams
    VectorXd t_gamma;
    for (int i = 0; i < p_trigrams.size() - 1; ++i)
    {
        if (i == 0)
        {
            PowerTrigram* lr_trigram = new PowerFullTrigram(t_counts, *(tri_disc_mats[i]), *(tri_ungammas[i]));
            trigram_vec->push_back(lr_trigram);
            t_gamma = *(tri_ungammas[i]);
        }
        else
        {
            cout << "computing low rank trigram\n";
            PowerLRTrigram* lr_trigram = new PowerLRTrigram(t_counts, *(p_trigrams[i]), *(tri_disc_mats[i]),
                t_gamma, *(tri_ungammas[i]), -1, tri_fvec[i - 1]);
            cout << "done computing low rank trigram\n";
            trigram_vec->push_back(lr_trigram);

            t_gamma = *(tri_ungammas[i]);
        }

    }

    // create the bigrams
    SparseMatrix<double>* b_counts = new SparseMatrix<double>();
    ComputeBCounts(*b_counts, *t_counts, *(p_trigrams[p_trigrams.size() - 1]));

    for (int i = 0; i < p_trigrams.size(); ++i)
        delete(p_trigrams[i]);

    for (int i = 0; i < tri_disc_mats.size(); ++i)
        delete(tri_disc_mats[i]);

    for (int i = 0; i < tri_ungammas.size(); ++i)
        delete(tri_ungammas[i]);

    vector< SparseMatrix<double>* > p_bigrams;
    vector< SparseMatrix<double>* > bi_disc_mats;
    vector < VectorXd* > bi_ungammas;

    p_bigrams.push_back(b_counts);
    // get powered/discounted bigrams
    for (int i = 0; i < tri_pvec.size(); ++i)
    {
        SparseMatrix<double>* p_bigram = new SparseMatrix<double>();
        SparseMatrix<double>* bi_disc_mat = new SparseMatrix<double>();
        VectorXd* bi_ungamma = new VectorXd();

        EigenHelper::Power(*b_counts, *p_bigram, bi_pvec[i]);
        double prev_P = 1.0;
        if (i > 0)
            prev_P = bi_pvec[i - 1];

        DiscountComputer::ComputeDiscount(*bi_disc_mat, *bi_ungamma, *(p_bigrams[i]), *p_bigram,
            prev_P, bi_dvec);

        p_bigrams.push_back(p_bigram);
        bi_disc_mats.push_back(bi_disc_mat);
        bi_ungammas.push_back(bi_ungamma);
    }

    SparseMatrix<double>* zero_bdisc = new SparseMatrix<double>(V, V);
    VectorXd* zero_gamma = new VectorXd(V);
    zero_gamma->fill(0);

    bi_disc_mats.push_back(zero_bdisc);
    bi_ungammas.push_back(zero_gamma);
    VectorXd b_gamma;
    // get low rank bigrams
    for (int i = 0; i < p_bigrams.size(); ++i)
    {
        if (i == 0)
        {
            PowerBigram* lr_bigram = new PowerFullBigram(*(p_bigrams[i]), *(bi_disc_mats[i]), *(bi_ungammas[i]));
            bigram_vec->push_back(lr_bigram);
            b_gamma = *(bi_ungammas[i]);
        }
        else
        {
            cout << "computing low rank bigram\n";
            PowerBigram* lr_bigram = new PowerLRBigram(*(p_bigrams[i]), *(bi_disc_mats[i]), b_gamma,
                *(bi_ungammas[i]), bi_fvec[i - 1]);
            cout << "done computing low rank bigram\n";
            bigram_vec->push_back(lr_bigram);

            b_gamma = *(bi_ungammas[i]);
        }
    }

    // unigrams (rank 1, power 0 bigram)
    VectorXd u_counts;
    ComputeUCounts(u_counts, *(p_bigrams[p_bigrams.size() - 1]));
    ComputeSmoothedUnigram(u_counts, uni_D);


    // delete intermediate quantities
    for (int i = 0; i < p_bigrams.size(); ++i)
        delete(p_bigrams[i]);

    for (int i = 0; i < bi_disc_mats.size(); ++i)
        delete(bi_disc_mats[i]);

    for (int i = 0; i < bi_ungammas.size(); ++i)
        delete(bi_ungammas[i]);
}



double QuadPLRE::GetCondProb(int i, int j, int k, int l)
{
    vector<double> vals;
    vals.reserve(7);
    vector<double> weights;
    weights.reserve(7);
    weights.push_back(1);


    for (int n = 0; n < quadgram_vec->size(); ++n)
    {
        double val = quadgram_vec->at(n)->CondLRDiscAt(i, j, k, l);
        vals.push_back(val);

        double curr_w = weights[n];
        double new_w = curr_w * quadgram_vec->at(n)->GammaAt(j, k, l);
        weights.push_back(new_w);
    }

    for (int n = 0; n < trigram_vec->size(); ++n)
    {
        double val = trigram_vec->at(n)->CondLRDiscAt(i, j, k);
        vals.push_back(val);

        double curr_w = weights[weights.size() - 1];
        double new_w = curr_w * trigram_vec->at(n)->GammaAt(j, k);
        weights.push_back(new_w);
    }

    for (int n = 0; n < bigram_vec->size(); ++n)
    {
        if (n < bigram_vec->size() - 1)
        {
            double val = bigram_vec->at(n)->CondLRDiscAt(i, j);
            vals.push_back(val);

            double curr_w = weights[weights.size() - 1];
            double new_w = curr_w * bigram_vec->at(n)->GammaAt(j);
            weights.push_back(new_w);
        }
        else
        {
            double val = smoothed_unigram->coeff(i);
            vals.push_back(val);
        }
    }

    assert(vals.size() == weights.size());

    double total_val = 0;
    string val_str = "";
    string w_str = "";
    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }
    assert(total_val >= 0 && total_val <= (1 + 1e-10));

    return total_val;
}




double QuadPLRE::GetQTCondProb(int i, int j, int k, int l)
{
    vector<double> vals;
    vector<double> weights;

    weights.push_back(1);


    for (int n = 0; n < quadgram_vec->size(); ++n)
    {
        double val = quadgram_vec->at(n)->CondLRDiscAt(i, j, k, l);
        vals.push_back(val);

        double curr_w = weights[weights.size() - 1];
        double new_w = curr_w * quadgram_vec->at(n)->GammaAt(j, k, l);
        weights.push_back(new_w);
    }

    for (int n = 0; n < 1; ++n)
    {
        double val = trigram_vec->at(n)->CondLRDiscAt(i, j, k);
        vals.push_back(val);

        double curr_w = weights[weights.size() - 1];
        double new_w = curr_w * trigram_vec->at(n)->GammaAt(j, k);
    }

    double total_val = 0;
    string val_str = "";
    string w_str = "";
    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }

    assert(total_val >= 0 && total_val <= (1 + 1e-10));

    return total_val;
}


double QuadPLRE::GetQTBCondProb(int i, int j, int k, int l)
{
    vector<double> vals;
    vector<double> weights;

    weights.push_back(1);


    for (int n = 0; n < quadgram_vec->size(); ++n)
    {
        double val = quadgram_vec->at(n)->CondLRDiscAt(i, j, k, l);
        vals.push_back(val);

        double curr_w = weights[n];
        double new_w = curr_w * quadgram_vec->at(n)->GammaAt(j, k, l);
        weights.push_back(new_w);
    }

    for (int n = 0; n < trigram_vec->size(); ++n)
    {
        double val = trigram_vec->at(n)->CondLRDiscAt(i, j, k);
        vals.push_back(val);

        double curr_w = weights[weights.size() - 1];
        double new_w = curr_w * trigram_vec->at(n)->GammaAt(j, k);
        weights.push_back(new_w);
    }

    for (int n = 0; n < 1; ++n)
    {
        double val = bigram_vec->at(n)->CondLRDiscAt(i, j);
        vals.push_back(val);
    }

    assert(vals.size() == weights.size());

    double total_val = 0;
    string val_str = "";
    string w_str = "";
    for (int n = 0; n < vals.size(); ++n)
    {
        double val = vals[n];
        double w = weights[n];
        total_val += w * val;
    }

    // print vals and print weights
    assert(total_val >= 0 && total_val <= (1 + 1e-10));
    return total_val;
}



void QuadPLRE::ComputeUCounts(VectorXd& u_counts, SparseMatrix<double>& zero_counts)
{
    u_counts.resize(V);
    u_counts.fill(0);
    map<int, double> total_counts;

    for (int i = 0; i < zero_counts.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(zero_counts, i); it; ++it)
        {
            int i1 = it.row();
            int i2 = it.col();
            double val = it.value();

            u_counts(i1) += val;

        }
    }
}


void QuadPLRE::ComputeBCounts(SparseMatrix<double>& b_counts, SMat3Tensor<int>& trigram_counts,
    SparseMatrix<double>& zero_trigram)
{
    b_counts.resize(V, V);

    map<pair<int, int>, double> total_counts;

    int mode = trigram_counts.GetMode();
    assert(mode == 0);

    for (int i = 0; i < zero_trigram.outerSize(); i++)
    {
        for (SparseMatrix<double>::InnerIterator it(zero_trigram, i); it; ++it)
        {
            int i1 = it.row();
            int i2 = it.col();
            double val = it.value();
            vector<int> indices = trigram_counts.GetTensorIndex(i1, i2);

            pair<int, int> p(indices[0], indices[1]);

            if (total_counts.count(p) == 0)
                total_counts[p] = 0;

            total_counts[p] += val;

        }
    }

    map<pair<int, int>, double>::iterator iter;
    vector<triplet> triplets;
    for (iter = total_counts.begin(); iter != total_counts.end(); ++iter)
    {
        int r = iter->first.first;
        int c = iter->first.second;
        double val = iter->second;
        triplets.push_back(triplet(r, c, val));
    }

    b_counts.setFromTriplets(triplets.begin(), triplets.end());

}

void QuadPLRE::ComputeTCounts(SMat3Tensor<int>& t_counts, SMat4Tensor<int>& quadgram_counts,
    SparseMatrix<double>& zero_quadgram)
{
    vector<int> dims = VectorPlus::CreateTriple(V, V, V);
    t_counts.Initialize(dims, 0);

    map<itriple, double> total_counts;

    int mode = quadgram_counts.GetMode();
    assert(mode == 0);

    map<itriple, int>& triple_2_index = quadgram_counts.GetTriple2IndexMap();
    map<itriple, int>::iterator iter1;
    for (iter1 = triple_2_index.begin(); iter1 != triple_2_index.end(); ++iter1)
    {
        itriple old_t = iter1->first;
        int c = iter1->second;
        Eigen::SparseVector<double> vec = zero_quadgram.col(c);

        for (Eigen::SparseVector<double>::InnerIterator iter(vec); iter; ++iter)
        {
            int r = iter.row();
            double val = iter.value();

            itriple t(r, old_t.get<0>(), old_t.get<1>());

            if (total_counts.count(t) == 0)
                total_counts[t] = 0;

            total_counts[t] += val;

        }
    }

    map<itriple, double>::iterator iter;
    t_counts.StartFill(total_counts.size());
    for (iter = total_counts.begin(); iter != total_counts.end(); ++iter)
    {
        itriple v = iter->first;
        int i1 = v.get<0>();
        int i2 = v.get<1>();
        int i3 = v.get<2>();

        int val = (int)(iter->second);
        t_counts.Set(i1, i2, i3, val);
    }
    t_counts.EndFill();
}

void QuadPLRE::ComputeSmoothedUnigram(VectorXd& u_counts, double uni_D)
{
    int non_zero = 0;
    for (int i = 0; i < u_counts.size(); ++i)
    {
        if (u_counts(i) > 0)
            non_zero += 1;
    }

    double sum = u_counts.sum();
    double gamma = non_zero * uni_D / sum;

    //	cout << "\t" << uni_D << "\t" << sum << "\t" << gamma << "\n";

    assert(gamma <= 1);

    smoothed_unigram = new VectorXd(V);
    smoothed_unigram->fill(0);

    for (int i = 0; i < u_counts.size(); ++i)
    {
        double val = u_counts.coeff(i);

        if (val > 0)
        {
            assert(val >= 1);
            (*smoothed_unigram)(i) = (val - uni_D) / sum + gamma / V;
        }
        else
        {
            (*smoothed_unigram)(i) = gamma / V;
        }
    }
    double smooth_sum = smoothed_unigram->sum();
    assert(abs(smooth_sum - 1.0) < 1e-8);
}
