/* Ankur Parikh
  QuadPLRE.hpp

  learns 4th order PLRE model */


#pragma once 

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/serialization/export.hpp>

#include <limits>
#include <vector>
#include <set>

#include "PowerLRTrigram.hpp"
#include "PowerFullTrigram.hpp"
#include "PowerTrigram.hpp"

#include "PowerQuadgram.hpp"
#include "PowerFullQuadgram.hpp"
#include "PowerLRQuadgram.hpp"

#include "eigen_boost_serialization.hpp"

#include "PowerBigram.hpp"
#include "PowerLRBigram.hpp"
#include "PowerFullBigram.hpp"



using namespace std;
using namespace Eigen;

class QuadPLRE
{

private:

    SpecMoments* spec_mom;
    VectorXd* smoothed_unigram;

    // ranks, rank = V / f
    vector<double>* quad_fvec;
    vector<double>* tri_fvec;
    vector<double>* bi_fvec;

    // powers
    vector<double>* quad_pvec;
    vector<double>* tri_pvec;
    vector<double>* bi_pvec;

    // discounts
    vector<double>* quad_dvec;
    vector<double>* tri_dvec;
    vector<double>* bi_dvec;

    // vocabulary
    int V;

    vector<PowerBigram*>* bigram_vec; // full/low rank bigrams
    vector<PowerTrigram*>* trigram_vec; // full/low rank trigrams
    vector<PowerQuadgram*>* quadgram_vec; // full/low rank quadgrams

    SMat3Tensor<int>* t_counts;


    friend class boost::serialization::access;


    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & spec_mom;
        ar & smoothed_unigram;
        ar & quad_fvec;
        ar & tri_fvec;
        ar & bi_fvec;

        ar & quad_pvec;
        ar & tri_pvec;
        ar & bi_pvec;

        ar & quad_dvec;
        ar & tri_dvec;
        ar & bi_dvec;

        ar & V;

        int bi_size = bigram_vec->size();
        ar & bi_size;

        PowerFullBigram* pfb = (PowerFullBigram*)(bigram_vec->at(0));
        ar & pfb;
        for (int i = 1; i < bigram_vec->size(); ++i)
        {
            PowerLRBigram* plrb = (PowerLRBigram*)(bigram_vec->at(i));
            ar & plrb;
        }

        int tri_size = trigram_vec->size();
        ar & tri_size;

        PowerFullTrigram* pft = (PowerFullTrigram*)(trigram_vec->at(0));
        ar & pft;
        for (int i = 1; i < trigram_vec->size(); ++i)
        {
            PowerLRTrigram* plrt = (PowerLRTrigram*)(trigram_vec->at(i));
            ar & plrt;
        }

        int quad_size = quadgram_vec->size();
        ar & quad_size;

        PowerFullQuadgram* pfq = (PowerFullQuadgram*)(quadgram_vec->at(0));
        ar & pfq;
        for (int i = 1; i < quadgram_vec->size(); ++i)
        {
            PowerLRQuadgram* plrq = (PowerLRQuadgram*)(quadgram_vec->at(i));
            ar & plrq;
        }

        ar & t_counts;
    }
    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        ar & spec_mom;
        ar & smoothed_unigram;
        ar & quad_fvec;
        ar & tri_fvec;
        ar & bi_fvec;

        ar & quad_pvec;
        ar & tri_pvec;
        ar & bi_pvec;

        ar & quad_dvec;
        ar & tri_dvec;
        ar & bi_dvec;

        ar & V;

        bigram_vec = new vector<PowerBigram*>();
        trigram_vec = new vector<PowerTrigram*>();
        quadgram_vec = new vector<PowerQuadgram*>();

        int bi_size;
        int tri_size;
        int quad_size;

        ar & bi_size;

        for (int i = 0; i < bi_size; ++i)
        {
            if (i == 0)
            {
                PowerFullBigram* pfb;
                ar & pfb;
                bigram_vec->push_back(pfb);
            }
            else if (i < bi_size - 1)
            {
                PowerLRBigram* plrb;
                ar & plrb;
                bigram_vec->push_back(plrb);
            }
            else
            {
                PowerLRBigram* plrb;
                ar & plrb;
                delete(plrb);
                bigram_vec->push_back(new PowerLRBigram());
            }
        }

        ar & tri_size;

        for (int i = 0; i < tri_size; ++i)
        {
            if (i == 0)
            {
                PowerFullTrigram* pft;
                ar & pft;
                trigram_vec->push_back(pft);
            }
            else
            {
                PowerLRTrigram* plrt;
                ar & plrt;
                trigram_vec->push_back(plrt);
            }
        }

        ar & quad_size;

        for (int i = 0; i < quad_size; ++i)
        {
            if (i == 0)
            {
                PowerFullQuadgram* pfq;
                ar & pfq;
                quadgram_vec->push_back(pfq);
            }
            else
            {
                PowerLRQuadgram* plrq;
                ar & plrq;
                quadgram_vec->push_back(plrq);
            }
        }

        ar & t_counts;
    }
    BOOST_SERIALIZATION_SPLIT_MEMBER()

public:

    QuadPLRE()
    {
        spec_mom = NULL;

        quad_fvec = NULL;
        tri_fvec = NULL;
        bi_fvec = NULL;

        quad_pvec = NULL;
        tri_pvec = NULL;
        bi_pvec = NULL;

        quad_dvec = NULL;
        tri_dvec = NULL;
        bi_dvec = NULL;

        bigram_vec = NULL;
        trigram_vec = NULL;

        t_counts = NULL;
    }

    QuadPLRE(SpecMoments* spec_mom,
        vector<double> quad_fvec, vector<double> tri_fvec, vector<double> bi_fvec,
        vector<double> quad_pvec, vector<double> tri_pvec, vector<double> bi_pvec,
        vector<double> quad_dvec, vector<double> tri_dvec, vector<double> bi_dvec, double uni_D);


    ~QuadPLRE()
    {
        delete(smoothed_unigram);
        delete(quad_fvec);
        delete(tri_fvec);
        delete(bi_fvec);
        delete(quad_pvec);
        delete(tri_pvec);
        delete(bi_pvec);
        delete(quad_dvec);
        delete(tri_dvec);
        delete(bi_dvec);

        for (int i = 0; i < bigram_vec->size(); ++i)
            delete(bigram_vec->at(i));
        for (int i = 0; i < trigram_vec->size(); ++i)
            delete(trigram_vec->at(i));
        for (int i = 0; i < quadgram_vec->size(); ++i)
            delete(quadgram_vec->at(i));

        delete(bigram_vec);
        delete(trigram_vec);
        delete(quadgram_vec);
        delete(t_counts);
    }

    double GetCondProb(int i, int j, int k, int l);

    double GetQTCondProb(int i, int j, int k, int l);

    double GetQTBCondProb(int i, int j, int k, int l);

    void ComputeTCounts(SMat3Tensor<int>& t_counts, SMat4Tensor<int>& trigram_counts,
        SparseMatrix<double>& zero_quadgram);

    void ComputeBCounts(SparseMatrix<double>& b_counts, SMat3Tensor<int>& trigram_counts,
        SparseMatrix<double>& zero_trigram);

    void ComputeUCounts(VectorXd& u_counts, SparseMatrix<double>& bigram_counts);

    void ComputeSmoothedUnigram(VectorXd& u_counts, double uni_D);

};
