/* Ankur Parikh
   SMat3Tensor.hpp

   matricized third order tensor, where one index is on the rows and 2 are on the columns
   */


#define EIGEN_YES_I_KNOW_SPARSE_MODULE_IS_NOT_STABLE_YET

# pragma once

#include <vector>
#include "Tensor.hpp"
#include <Eigen/Sparse>
#include "Tensor.hpp"
#include <map>
#include <iostream>
#include <fstream>
#include "eigen_boost_serialization.hpp"

using namespace std;
using namespace Eigen;

// complex tensor
template <class T>
class SMat3Tensor
{
private:

    SparseMatrix<T>* sp_ten_alt; //sparse matrix

    map<pair<int, int>, int>* pair_to_index; // tensor index pair to column index
    map<int, pair<int, int> >* index_to_pair; // column index to tensor index pair 
    vector<Triplet<T> >* alt_triplet_list; // sparse matrix in triplet form
    vector<int>* dims; // dimensions of tensor

    int curr_sub_index;

    bool isInitialized; // flag
    bool isFilled; // if tensor is filled or not

    int curr_alt_index;

    int mode; // the mode that is on the rows

    void ComputeOffsets();

    // Used for indexing
    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & sp_ten_alt;
        ar & pair_to_index;
        ar & index_to_pair;
        //	ar & alt_triplet_list;
        ar & dims;
        ar & curr_sub_index;
        ar & isInitialized;
        ar & isFilled;
        ar & curr_alt_index;
        ar & mode;
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        ar & sp_ten_alt;
        ar & pair_to_index;
        ar & index_to_pair;
        delete(index_to_pair);
        index_to_pair = new map<int, pair<int, int> >();
        //	ar & alt_triplet_list;
        ar & dims;
        ar & curr_sub_index;
        ar & isInitialized;
        ar & isFilled;
        ar & curr_alt_index;
        ar & mode;
    }



    BOOST_SERIALIZATION_SPLIT_MEMBER()


public:

    SMat3Tensor();
    SMat3Tensor(vector<int>& dim_array, int mode);
    ~SMat3Tensor();

    // intialization
    void Initialize(vector<int>& dim_array, int mode);

    void ResetSpMat()
    {
        delete(sp_ten_alt);
        sp_ten_alt = new SparseMatrix<T>(1, 1);
    }

    int GetMode() { return mode; }

    // basic accessors
    vector<int>& Dims() { return *dims; }
    int Dim(int index) { return (*dims)[index]; }

    // access elements of tensor
    double At(int i, int j, int k);

    void StartFill(int reserve_count);
    void StartFill();
    void EndFill();

    // set elements of tensor
    void Set(int i, int j, int k, T val);

    SparseMatrix<T>& GetMat() { return *sp_ten_alt; }

    int ComputeAltIndex(int i, int k);

    vector<int> GetTensorIndex(int i, int j);

    pair<int, int> ComputeReverseAltIndex(int i);
};


// default constructor
template <class T>
SMat3Tensor<T>::SMat3Tensor()
{
    sp_ten_alt = NULL;

    pair_to_index = NULL;
    index_to_pair = NULL;
    alt_triplet_list = NULL;
    dims = NULL; // dimensions of tensor


    isInitialized = false; // flag
    isFilled = false;

    curr_alt_index = -1;

    mode = -1;
}

template <class T>
SMat3Tensor<T>::SMat3Tensor(vector<int>& dim_array, int mode)
{
    alt_triplet_list = NULL;
    isFilled = false;
    isInitialized = false;
    Initialize(dim_array, mode);
}

// destructor
template <class T>
SMat3Tensor<T>::~SMat3Tensor()
{
    delete(sp_ten_alt);
    delete(pair_to_index);
    delete(index_to_pair);
    //	delete(alt_triplet_list);
    delete(dims);
}

// initialize
template <class T>
void SMat3Tensor<T>::Initialize(vector<int>& dims, int mode)
{
    assert(!isInitialized);

    assert(dims.size() == 3);

    sp_ten_alt = new SparseMatrix<T>();
    pair_to_index = new map<pair<int, int>, int>();
    index_to_pair = new map<int, pair<int, int> >();
    this->dims = new vector<int>(dims);

    curr_alt_index = 0;

    isInitialized = true;
    this->mode = mode;
}


// access elements of tensor
template <class T>
double SMat3Tensor<T>::At(int i, int j, int k)
{
    assert(isInitialized);
    assert(isFilled);

    if (mode == 0)
    {
        assert(isFilled);
        int alt_index = ComputeAltIndex(j, k);
        if (alt_index < 0)
            return 0;
        else
            return sp_ten_alt->coeff(i, alt_index);
    }
    else if (mode == 2)
    {
        int alt_index = ComputeAltIndex(i, j);
        if (alt_index < 0)
            return 0;
        else
            return sp_ten_alt->coeff(alt_index, k);
    }
    else
    {
        assert(0);
    }
}

template <class T>
void SMat3Tensor<T>::StartFill(int reserve_count)
{
    assert(!isFilled);
    alt_triplet_list = new vector< Triplet<T> >();
    alt_triplet_list->reserve(reserve_count);
}

template <class T>
void SMat3Tensor<T>::StartFill()
{
    assert(!isFilled);
    alt_triplet_list = new vector< Triplet<T> >();
}

template <class T>
void SMat3Tensor<T>::EndFill()
{
    if (mode == 0)
    {
        sp_ten_alt->resize(dims->at(0), curr_alt_index);
    }
    else if (mode == 2)
    {
        sp_ten_alt->resize(curr_alt_index, dims->at(2));
    }

    sp_ten_alt->setFromTriplets(alt_triplet_list->begin(), alt_triplet_list->end());
    isFilled = true;
    delete(alt_triplet_list);
}


template <class T>
void SMat3Tensor<T>::Set(int i, int j, int k, T val)
{
    assert(!isFilled);
    assert(alt_triplet_list != NULL);

    if (mode == 0)
    {
        int alt_col_index = ComputeAltIndex(j, k);
        if (alt_col_index == -1)
        {
            pair<int, int> p(j, k);
            (*pair_to_index)[p] = curr_alt_index;
            (*index_to_pair)[curr_alt_index] = p;
            curr_alt_index++;
        }
        alt_col_index = ComputeAltIndex(j, k);
        assert(alt_col_index >= 0);

        alt_triplet_list->push_back(Triplet<T>(i, alt_col_index, val));
    }
    else if (mode == 2)
    {
        int alt_col_index = ComputeAltIndex(i, j);
        if (alt_col_index == -1)
        {
            pair<int, int> p(i, j);
            (*pair_to_index)[p] = curr_alt_index;
            (*index_to_pair)[curr_alt_index] = p;
            curr_alt_index++;
        }
        alt_col_index = ComputeAltIndex(i, j);
        assert(alt_col_index >= 0);

        alt_triplet_list->push_back(Triplet<T>(alt_col_index, k, val));
    }
    else
    {
        assert(0);
    }
}

template <class T>
int SMat3Tensor<T>::ComputeAltIndex(int i, int j)
{
    pair<int, int> p(i, j);
    if (pair_to_index->count(p) > 0)
        return (*pair_to_index)[p];
    else
        return -1;
}

// get original tensor index pair from column index i
template <class T>
pair<int, int> SMat3Tensor<T>::ComputeReverseAltIndex(int i)
{
    if (index_to_pair->count(i) > 0)
        return (*index_to_pair)[i];
    else
        return pair<int, int>(-1, -1);
}

// get original tensor indices(i,j,k) from matrix indices (i,j)
template <class T>
vector<int> SMat3Tensor<T>::GetTensorIndex(int i, int j)
{

    if (mode == 0)
    {
        pair<int, int> orig_indices = ComputeReverseAltIndex(j);
        vector<int> v = VectorPlus::CreateTriple(i, orig_indices.first, orig_indices.second);
        return v;
    }
    else if (mode == 2)
    {
        pair<int, int> orig_indices = ComputeReverseAltIndex(i);
        vector<int> v = VectorPlus::CreateTriple(orig_indices.first, orig_indices.second, j);
        return v;
    }
    else
    {
        assert(0);
    }
}