/* Ankur Parikh
   SQuadMatrixArray.hpp

   a certain storage format for a 3rd order tensor
   basically a vector of sparse matrices,

   e.g. if mode = 1 (starting from zero) then

   each index i in the vector stores the tensor slice T(:,i,:)
   */

#pragma once

#include "Tensor.hpp"
#include "VectorPlus.hpp"
#include <Eigen/Sparse>
#include "CondensedSMat.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;


template <class T>
class SMatrixArray : public Tensor < T > {

private:

    int mode; // mode to index on
    vector<int>* dims; // dimensions of tensor

    // vector of sparse matrices, each matrix represented implicitly as vector of triplets
    vector < vector< Triplet<T> >* >* triplet_list_vec;

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & mode;
        ar & dims;
        ar & triplet_list_vec;
    }

public:

    SMatrixArray()
    {
        dims = NULL;
        triplet_list_vec = NULL;

        mode = -1;
    }

    // constructors, initializes all values to zero
    void Initialize(vector<int>& dims, int mode)
    {
        this->dims = new vector<int>(dims);
        this->mode = mode;
    }

    ~SMatrixArray()
    {
        delete(dims);

        for (int i = 0; i < triplet_list_vec->size(); ++i)
        {
            delete(triplet_list_vec->at(i));
        }

        delete(triplet_list_vec);
    }

    // basic accessors
    vector<int>& Dims() { return *dims; }
    int Dim(int index) { return (*dims)[index]; }

    void startFill()
    {
        triplet_list_vec = new vector < vector< Triplet<T> >* >(dims->at(mode), NULL);
        for (int i = 0; i < triplet_list_vec->size(); ++i)
        {
            triplet_list_vec->at(i) = new vector< Triplet<T> >();
        }
    }

    void endFill()
    {

    }

    void Set(int i, int j, int k, T val)
    {
        if (mode == 0)
        {
            triplet_list_vec->at(i)->push_back(Triplet<T>(j, k, val));
        }
        else if (mode == 1)
        {
            triplet_list_vec->at(j)->push_back(Triplet<T>(i, k, val));
        }
        else if (mode == 2)
        {
            triplet_list_vec->at(k)->push_back(Triplet<T>(i, j, val));
        }
        else
        {
            assert(0);
        }
    }

    void GetMat(SparseMatrix<T, RowMajor>& sp_mat, int mode, int index)
    {
        assert(this->mode == mode);
        if (mode == 0)
        {
            sp_mat.resize(dims->at(1), dims->at(2));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else if (mode == 1)
        {
            sp_mat.resize(dims->at(0), dims->at(2));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else if (mode == 2)
        {
            sp_mat.resize(dims->at(0), dims->at(1));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else
        {
            assert(0);
        }
    }

    void GetMat(SparseMatrix<T>& sp_mat, int mode, int index)
    {
        assert(this->mode == mode);
        if (mode == 0)
        {
            sp_mat.resize(dims->at(1), dims->at(2));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else if (mode == 1)
        {
            sp_mat.resize(dims->at(0), dims->at(2));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else if (mode == 2)
        {
            sp_mat.resize(dims->at(0), dims->at(1));
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else
        {
            assert(0);
        }
    }

    void GetMat(CondensedSMat& sp_mat, int mode, int index)
    {
        assert(this->mode == mode);
        sp_mat.Initialize(*(triplet_list_vec->at(index)), dims->at(0));
    }

    // access elements of tensor, shouldn't be called
    T At(int i, int j, int k)
    {
        assert(0);
        return 0;
        /*	if (mode == 1)
            {
            return m_array->at(i)->coeff(j,k);
            }
            else if (mode == 2)
            {
            return m_array->at(j)->coeff(i,k);
            }
            else if (mode == 3)
            {
            return m_array->at(k)->coeff(i,j);
            }
            else
            {
            assert(0);
            } */
    }
};