/* Ankur Parikh
   SQuadMatrixArray.hpp

   a certain storage format for a 4th order tensor
   basically a vector of sparse matrices,

   e.g. if mode1 = 1, mode2 = 2 (starting from zero) then

   each index i in the vector  maps to two indices (j1, j2) so that the matrix stored at position i
   is the slice T(:,j1,j2,:) of the tensor T

   */


#pragma once

#include "Tensor.hpp"
#include "VectorPlus.hpp"
#include <Eigen/Sparse>
#include "CondensedSMat.hpp"
#include "eigen_boost_serialization.hpp"

using namespace Eigen;
using namespace std;


template <class T>
class SQuadMatrixArray  {

private:

    map<pair<int, int>, int>* pair_to_index; // map mode pairs to index in vector
    map<int, pair<int, int> >* index_to_pair; // map index in vector to mode pairs

    int mode1; // mode 1
    int mode2; // mode 2
    int curr_index;

    vector<int>* dims; // dimensions of tensor

    vector < vector< Triplet<T> >* >* triplet_list_vec; // store all matrices implicitly as triplet vectors

    friend class boost::serialization::access;

    template<typename Archive>
    void serialize(Archive& ar, const unsigned version) {

        ar & pair_to_index;
        ar & index_to_pair;
        ar & mode1;
        ar & mode2;
        ar & curr_index;
        ar & dims;
        ar & triplet_list_vec;
    }

public:

    SQuadMatrixArray()
    {
        dims = NULL;
        triplet_list_vec = NULL;
        pair_to_index = NULL;
        index_to_pair = NULL;
        mode1 = -1;
        mode2 = -1;
        curr_index = -1;
    }

    // constructors, initializes all values to zero
    void Initialize(vector<int>& dims, int mode1, int mode2)
    {
        this->dims = new vector<int>(dims);
        curr_index = 0;
        pair_to_index = new map<pair<int, int>, int>();
        index_to_pair = new map<int, pair<int, int> >();
        this->mode1 = mode1;
        this->mode2 = mode2;
        assert(mode1 == 1 && mode2 == 2);
    }

    ~SQuadMatrixArray()
    {
        delete(dims);

        for (int i = 0; i < triplet_list_vec->size(); ++i)
        {
            delete(triplet_list_vec->at(i));
        }

        delete(triplet_list_vec);

        delete(pair_to_index);
        delete(index_to_pair);
    }

    // basic accessors
    vector<int>& Dims() { return *dims; }
    int Dim(int index) { return (*dims)[index]; }

    void startFill(int reserve_count)
    {
        triplet_list_vec = new vector < vector< Triplet<T> >* >();
        triplet_list_vec->reserve(reserve_count);
    }

    void startFill()
    {
        triplet_list_vec = new vector < vector< Triplet<T> >* >();
    }

    void endFill()
    {

    }

    // set value of tensor
    void Set(int i, int j, int k, int l, T val)
    {
        if (mode1 == 1 && mode2 == 2)
        {
            int index = GetPairIndex(j, k);
            triplet_list_vec->at(index)->push_back(Triplet<T>(i, l, val));
        }
        else
        {
            assert(0);
        }
    }

    int GetPairIndex(int j, int k)
    {
        pair<int, int> p(j, k);
        if (pair_to_index->count(p) == 0)
        {
            (*pair_to_index)[p] = curr_index;
            (*index_to_pair)[curr_index] = p;
            triplet_list_vec->push_back(new vector< Triplet<T> >());
            curr_index++;
        }
        return (*pair_to_index)[p];
    }

    int GetTestPairIndex(int j, int k)
    {
        pair<int, int> p(j, k);
        if (pair_to_index->count(p) == 0)
        {
            assert(0);
        }
        return (*pair_to_index)[p];
    }


    map<int, pair<int, int> >& GetIndexMap() { return *pair_to_index; }


    void GetMat(SparseMatrix<T, RowMajor>& sp_mat, int mode1, int mode2, int j, int k)
    {
        assert(this->mode1 = mode1);
        assert(this->mode2 = mode2);

        if (mode1 == 1 && this->mode2 == 2)
        {
            sp_mat.resize(dims->at(0), dims->at(3));

            int index = GetTestPairIndex(j, k);
            sp_mat.setFromTriplets(triplet_list_vec->at(index)->begin(), triplet_list_vec->at(index)->end());
        }
        else
        {
            assert(0);
        }
    }


    void GetMat(CondensedSMat& sp_mat, int mode1, int mode2, int j, int k)
    {
        assert(this->mode1 = mode1);
        assert(this->mode2 = mode2);

        int index = GetTestPairIndex(j, k);
        sp_mat.Initialize(*(triplet_list_vec->at(index)), dims->at(0));
    }
};