/* Ankur Parikh
   SpecMoments.hpp

   class that computes/gives access to n-grams of various orders
   */

#pragma once 

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include "SMatrixArray.hpp"
#include <limits>
#include <vector>
#include <set>
#include "SMat3Tensor.hpp"
#include "SMat4Tensor.hpp"
#include "SQuadMatrixArray.hpp"
#include "eigen_boost_serialization.hpp"
#include <boost/unordered_map.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>
#include <boost/tuple/tuple_io.hpp>
#include "serialize_tuple.hpp"
#include <boost/tr1/unordered_map.hpp>

using namespace std;
using namespace Eigen;

typedef Triplet<double> triplet;


#define ASSERT_WITH_MESSAGE(condition, message) do { \
if (!(condition)) { printf((message)); } \
assert ((condition)); } while(false)



class SpecMoments {

private:

    map<string, int>* ngram_map; // word to index
    map<int, string>* rev_ngram_map; // index to word

    int num_cores;
    string oov_marker; // out of vocabulary marker

    // unigrams, bigrams, trigrams, quadgrams in vector/matrix form
    VectorXd* unigrams;
    SparseMatrix<int>* bigram_cmat;
    SMat3Tensor<int>* trigram_cmat;
    SMat4Tensor<int>* quadgram_cmat;

    //unigrams, bigrams, trigrams, quadgrams in dictionary form
    map<int, int>* unigram_counts;
    map<pair<int, int>, int>* bigram_counts;
    map<itriple, int>* trigram_counts;
    map<iquadple, int>* quadgram_counts;

    // unigram history, bigram history, trigram history counts in map form
    // e.g. uh_count["fun"] = number of times fun precedes a word in a corpus
    map<itriple, int>* th_counts;
    map<pair<int, int>, int>* bh_counts;
    map<int, int>* uh_counts;

    int curr_w_index;
    int num_sents;

    // boost stuff


    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        ar & ngram_map;
        ar & rev_ngram_map;
        ar & unigrams;
        ar & unigram_counts;
        ar & bigram_counts;
        ar & trigram_counts;
        ar & quadgram_counts;

        ar & curr_w_index;
        ar & num_sents;

        ar & oov_marker;

        ar & bigram_counts;
        ar & trigram_counts;
        ar & quadgram_counts;

        ar & th_counts;
        ar & bh_counts;
        ar & uh_counts;
    }

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
    {
        ar & ngram_map;
        ar & rev_ngram_map;
        ar & unigrams;
        ar & unigram_counts;
        ar & bigram_counts;
        ar & trigram_counts;
        ar & quadgram_counts;

        ar & curr_w_index;
        ar & num_sents;

        ar & oov_marker;

        ar & bigram_counts;
        ar & trigram_counts;
        ar & quadgram_counts;

        ar & th_counts;
        ar & bh_counts;
        ar & uh_counts;

        delete(rev_ngram_map);
        rev_ngram_map = new map<int, string>();
    }



    BOOST_SERIALIZATION_SPLIT_MEMBER()


public:

    SpecMoments()
    {
        ngram_map = NULL;
        rev_ngram_map = NULL;
        num_cores = -1;
        oov_marker = "";

        unigrams = NULL;
        bigram_cmat = NULL;
        trigram_cmat = NULL;
        quadgram_cmat = NULL;

        unigram_counts = NULL;
        bigram_counts = NULL;
        trigram_counts = NULL;
        quadgram_counts = NULL;

        th_counts = NULL;
        bh_counts = NULL;
        uh_counts = NULL;

        curr_w_index = -1;
        num_sents = -1;
    }


    SpecMoments(string fn, string oov_arg,
        string start_marker, string end_marker, int cores = 1);

    ~SpecMoments()
    {
        delete(ngram_map);
        delete(rev_ngram_map);
        delete(unigrams);
        delete(unigram_counts);
        delete(bigram_counts);
        delete(trigram_counts);
        delete(quadgram_counts);

        DeleteTensors();
        DeleteBHMaps();
    }

    void readUnigrams(string fn, string start_marker, string end_marker);
    void readBigrams(string fn);
    void readMatTrigrams(string fn);
    void readMatQuadgrams(string fn);

    void FillHCounts();

    void FillTensors();

    //  void ClearSlowMap() { delete(ngram_map); ngram_map = new map<string,int>(); }

    // delete tensors since they take up space
    void DeleteTensors()
    {
        if (bigram_cmat != NULL)
            delete(bigram_cmat);

        if (trigram_cmat != NULL)
            delete(trigram_cmat);

        if (quadgram_cmat != NULL)
            delete(quadgram_cmat);
    }

    // delete history maps since they take up space
    void DeleteBHMaps()
    {
        if (th_counts != NULL)
            delete(th_counts);
        if (bh_counts != NULL)
            delete(bh_counts);
        if (uh_counts != NULL)
            delete(uh_counts);
    }

    void ResetCountMaps()
    {
        delete(unigram_counts);
        delete(bigram_counts);
        delete(trigram_counts);
        delete(quadgram_counts);

        unigram_counts = new  map<int, int>();
        bigram_counts = new map<pair<int, int>, int>();
        trigram_counts = new map<itriple, int>();
        quadgram_counts = new map<iquadple, int>();
    }

    int GetV() { return curr_w_index; };



    string GetWordStr(int index)
    {
        return (*rev_ngram_map)[index];
    }


    //getters
    map<int, int>& getUnigramCounts() { return *unigram_counts; }
    map<pair<int, int>, int>& getBigramCounts() { return *bigram_counts; }


    map<itriple, int>& getTrigramCounts() { return *trigram_counts; }
    map<iquadple, int>& getQuadgramCounts() { return *quadgram_counts; }

    VectorXd& getUnigrams() { return *unigrams; }
    SparseMatrix<int>& getBigramCMat() { return *bigram_cmat; }

    // SMatrixArray<double>& getTrigrams() { return *trigrams; }
    SMat3Tensor<int>& getTrigramCMat() { return *trigram_cmat; }
    SMat4Tensor<int>& getQuadgramCMat() { return *quadgram_cmat; }

    int GetTotalBC()
    {
        map<pair<int, int>, int>::iterator iter;
        int total_count = 0;
        for (iter = bigram_counts->begin(); iter != bigram_counts->end(); ++iter)
        {
            total_count += iter->second;
        }

        return total_count;
    }

    int GetTotalTC()
    {
        map<itriple, int>::iterator iter;
        int total_count = 0;
        for (iter = trigram_counts->begin(); iter != trigram_counts->end(); ++iter)
        {
            total_count += iter->second;
        }

        return total_count;
    }

    int GetID(string word)
    {

        if (ngram_map->count(word) > 0)
            return (*ngram_map)[word];
        else
            return std::numeric_limits<int>::max();
    }

    // id of word during test time, basically if word is not found return the oov_marker id
    int GetFinalID(const string& word)
    {
        if (ngram_map->count(word) > 0)
            return (*ngram_map)[word];
        else
            return (*ngram_map)[oov_marker];
    }

    int GetOOVIndex()
    {
        return (*ngram_map)[oov_marker];
    }

    map<itriple, int>& GetTHCountMap() { return *th_counts; }

    map<pair<int, int>, int>& GetBHCountMap()
    {
        return *bh_counts;
    }

    map<int, int>& GetUHCountMap()
    {
        return *uh_counts;
    }

    int GetTHCount(itriple t)
    {
        if (th_counts->count(t) > 0)
        {
            return (*th_counts)[t];
        }
        else
        {
            return 0;
        }
    }

    int GetBHCount(pair<int, int> p)
    {
        if (bh_counts->count(p) > 0)
        {
            return (*bh_counts)[p];
        }
        else
        {
            return 0;
        }
    }

    int GetUHCount(int i)
    {
        if (uh_counts->count(i) > 0)
        {
            return (*uh_counts)[i];
        }
        else
        {
            return 0;
        }
    }



    double GetCondProb(int i, int j); // P(i|j) according to empirical bigram
    double GetCondProb(int i, int j, int k); // P(i|j, k) according to empirical trigram
};
