/* Ankur Parikh
  run_plre.cpp

  The main file for training/testing PLRE. See README for details */

#include "Tensor.hpp"
# include <string>
#include "IOLibrary.hpp"
#include "VectorPlus.hpp"
#include "SMatrixArray.hpp"
#include "SpecMoments.hpp"
#include "IWittenBell.hpp"
#include "IKneserNey.hpp"
#include "EM.hpp"
#include "QuadPLRE.hpp"
#include "PLRE.hpp"
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include "eigen_boost_serialization.hpp"
#include <boost/serialization/export.hpp>
#include <boost/algorithm/string.hpp>
#include "QuadLMWrapper.hpp"
#include "omp.h"

using namespace Eigen;

int main(int argc, char *argv[])
{
    // these are the defaults if no params are set
    string model_fn = "plre-model";
    int order = 0;
    double quad_f1 = 0;
    double tri_f1 = 0;
    double bi_f1 = 0;
    double P = 0.5;

    string train_fn = "";
    string test_fn = "";

    string oov_marker = "<*oov*>";
    string start_marker = "<s>";
    string end_marker = "</s>";


    for (int i = 1; i < argc; i = i + 2)
    {
        string arg_type = argv[i];
        string arg_val = argv[i + 1];

        if (arg_type == "--order")
        {
            order = atoi(arg_val.c_str());
        }
        else if (arg_type == "--qrank")
        {
            quad_f1 = atof(arg_val.c_str());
        }
        else if (arg_type == "--trank")
        {
            tri_f1 = atof(arg_val.c_str());
        }
        else if (arg_type == "--brank")
        {
            bi_f1 = atof(arg_val.c_str());
        }
        else if (arg_type == "--model")
        {
            model_fn = arg_val;
        }
        else if (arg_type == "--train")
        {
            train_fn = arg_val;
        }
        else if (arg_type == "--test")
        {
            test_fn = arg_val;
        }
        else if (arg_type == "--startmarker")
        {
            start_marker = arg_val;
        }
        else if (arg_type == "--endmarker")
        {
            end_marker = arg_val;
        }
        else if (arg_type == "--oovmarker")
        {
            oov_marker = arg_val;
        }
        else
        {
            ASSERT_WITH_MESSAGE(0, "command line argument not valid");
        }
    }

    /*order = 4;
    quad_f1 = 0.001;
    tri_f1 = 0.001;
    bi_f1 = 0.001;
    train_fn = "train-sm.txt";
    test_fn = "test-sm.txt";
    oov_marker = "the";*/


    cout << "\n parameter values \n";
    cout << "model name: " << model_fn << "\n";
    cout << "qrank: " << quad_f1 << "\n";
    cout << "trank: " << tri_f1 << "\n";
    cout << "brank: " << bi_f1 << "\n";
    cout << "train file: " << train_fn << "\n";
    cout << "test file: " << test_fn << "\n";
    cout << "oov marker: " << oov_marker << "\n";
    cout << "start marker: " << start_marker << "\n";
    cout << "end marker: " << end_marker << "\n";

    // check mandatory arguments are there
    ASSERT_WITH_MESSAGE(train_fn.size() > 0, "no training file provided");
    ASSERT_WITH_MESSAGE(order <= 4 && order >= 2, "max order must be between 2 and 4");
    ASSERT_WITH_MESSAGE(tri_f1 > 0 && bi_f1 > 0, "code requires low rank bigrams and trigrams");

    //assert(0);

    SpecMoments* moments = new SpecMoments(train_fn, oov_marker, start_marker, end_marker);


    moments->FillHCounts();

    cout << "\n\n--------Learning Kneser Ney models-------------\n";
    IWittenBell* wb1 = new IWittenBell(moments);
    IWittenBell* wb2 = new IWittenBell(moments);

    IKneserNey* kney = new IKneserNey(moments, wb1, true);
    IQuadKneserNey* qkney = new IQuadKneserNey(moments, wb2, true);
    cout << "---------Done learning Kneser Ney models----------\n";

    // use good turing estimates for discounts
    double q_gt_quadD1 = qkney->GetQD1();
    double q_gt_quadD2 = qkney->GetQD2();
    double q_gt_quadD3 = qkney->GetQD3();
    double q_gt_triD1 = qkney->GetTriD1();
    double q_gt_triD2 = qkney->GetTriD2();
    double q_gt_triD3 = qkney->GetTriD3();
    double q_gt_biD1 = qkney->GetBD1();
    double q_gt_biD2 = qkney->GetBD2();
    double q_gt_biD3 = qkney->GetBD3();

    double q_qD1 = pow(q_gt_quadD1, 0.5);
    double q_qD2 = pow(q_gt_quadD1, 0.5);
    double q_qD3 = pow(q_gt_quadD1, 0.5);

    double q_tD1 = pow(q_gt_triD1, 0.5);
    double q_tD2 = pow(q_gt_triD1, 0.5);
    double q_tD3 = pow(q_gt_triD1, 0.5);

    double q_bD1 = pow(q_gt_biD1, 0.5);
    double q_bD2 = pow(q_gt_biD1, 0.5);
    double q_bD3 = pow(q_gt_biD1, 0.5);

    // these are the discounts from the 3-gram Kneser Ney
    double t_gt_triD1 = kney->GetTriD1();
    double t_gt_triD2 = kney->GetTriD2();
    double t_gt_triD3 = kney->GetTriD3();

    double t_gt_biD1 = kney->GetBD1();
    double t_gt_biD2 = kney->GetBD2();
    double t_gt_biD3 = kney->GetBD3();

    double t_tD1 = pow(t_gt_triD1, 0.5);
    double t_tD2 = pow(t_gt_triD1, 0.5);
    double t_tD3 = pow(t_gt_triD1, 0.5);

    double t_bD1 = pow(t_gt_biD1, 0.5);
    double t_bD2 = pow(t_gt_biD1, 0.5);
    double t_bD3 = pow(t_gt_biD1, 0.5);

    // these are the discounts from the 2-gram Kneser Ney
    double b_gt_biD1 = kney->GetBiD1();
    double b_gt_biD2 = kney->GetBiD2();
    double b_gt_biD3 = kney->GetBiD3();


    double b_bD1 = pow(b_gt_biD1, 0.5);
    double b_bD2 = pow(b_gt_biD1, 0.5);
    double b_bD3 = pow(b_gt_biD1, 0.5);

    double uni_D = kney->GetUniD1();


    bool mod_flag = true;
    if (mod_flag == true)
    {
        q_qD2 = pow(q_gt_quadD2, 0.5);
        q_qD3 = pow(q_gt_quadD3, 0.5);

        q_tD2 = pow(q_gt_triD2, 0.5);
        q_tD3 = pow(q_gt_triD3, 0.5);

        q_bD2 = pow(q_gt_biD2, 0.5);
        q_bD3 = pow(q_gt_biD3, 0.5);

        t_tD2 = pow(t_gt_triD2, 0.5);
        t_tD3 = pow(t_gt_triD3, 0.5);

        t_bD2 = pow(t_gt_biD2, 0.5);
        t_bD3 = pow(t_gt_biD3, 0.5);

        b_bD2 = pow(b_gt_biD2, 0.5);
        b_bD3 = pow(b_gt_biD3, 0.5);
    }
    else
    {
        ASSERT_WITH_MESSAGE(0, "mod flag should always be true");
    }

    cout.flush();

    vector<double> q_quad_fvec;
    vector<double> q_tri_fvec;
    vector<double> q_bi_fvec;

    vector<double> q_quad_pvec;
    vector<double> q_tri_pvec;
    vector<double> q_bi_pvec;

    QuadPLRE* quad_plre = NULL;
    PLRE* tri_plre = NULL;
    PLRE* bi_plre = NULL;


    cout << "\n\n----------starting learning PLRE model----------\n";
    moments->FillTensors();
    double timer_start = omp_get_wtime();

    if (order == 4)
    {
        // quadgram PLRE model
        if (quad_f1 > 0)
        {
            q_quad_fvec = VectorPlus::CreatePair(quad_f1, 0.0);
            q_quad_pvec = VectorPlus::CreatePair(P, 0.0);
        }
        else
        {
            q_quad_fvec = VectorPlus::CreateSingleton(0.0);
            q_quad_pvec = VectorPlus::CreateSingleton(0.0);
        }

        q_tri_fvec = VectorPlus::CreatePair(tri_f1, 0.0);
        q_bi_fvec = VectorPlus::CreatePair(bi_f1, 0.0);

        q_tri_pvec = VectorPlus::CreatePair(P, 0.0);
        q_bi_pvec = VectorPlus::CreatePair(P, 0.0);

        vector<double> q_quad_dvec = VectorPlus::CreateTriple(q_qD1, q_qD2, q_qD3);
        vector<double> q_tri_dvec = VectorPlus::CreateTriple(q_tD1, q_tD2, q_tD3);
        vector<double> q_bi_dvec = VectorPlus::CreateTriple(q_bD1, q_bD2, q_bD3);

        cout << "\n\n----------starting to learn order 4 PLRE----------\n";
        quad_plre = new QuadPLRE(moments,
            q_quad_fvec, q_tri_fvec, q_bi_fvec,
            q_quad_pvec, q_tri_pvec, q_bi_pvec,
            q_quad_dvec, q_tri_dvec, q_bi_dvec, uni_D);
        cout << "----------finished learning order 4 PLRE----------\n";
    }

    if (order >= 3)
    {
        // learn trigram PLRE next
        vector<double> t_tri_fvec = VectorPlus::CreatePair(tri_f1, 0.0);
        vector<double> t_bi_fvec = VectorPlus::CreatePair(bi_f1, 0.0);

        vector<double> t_tri_pvec = VectorPlus::CreatePair(P, 0.0);
        vector<double> t_bi_pvec = VectorPlus::CreatePair(P, 0.0);

        vector<double> t_tri_dvec = VectorPlus::CreateTriple(t_tD1, t_tD2, t_tD3);
        vector<double> t_bi_dvec = VectorPlus::CreateTriple(t_bD1, t_bD2, t_bD3);

        cout << "\n\n----------starting to learn order 3 PLRE----------\n";
        tri_plre = new PLRE(moments,
            t_tri_fvec, t_bi_fvec,
            t_tri_pvec, t_bi_pvec,
            t_tri_dvec, t_bi_dvec, uni_D);
        cout << "----------finished learning order 3 PLRE----------\n";
    }

    // learn bigram PLRE last
    vector<double> b_bi_fvec = VectorPlus::CreatePair(bi_f1, 0.0);
    vector<double> b_bi_pvec = VectorPlus::CreatePair(P, 0.0);
    vector<double> b_bi_dvec = VectorPlus::CreateTriple(b_bD1, b_bD2, b_bD3);

    cout << "\n\n----------starting to learn order 2 PLRE----------\n";
    bi_plre = new PLRE(moments, b_bi_fvec, b_bi_pvec, b_bi_dvec, uni_D);
    cout << "----------starting to learn order 2 PLRE----------\n";

    cout << "\n\n----------done learning models----------" << "\n";
    cout.flush();

    double total_elapsed = omp_get_wtime();
    cout << "PLRE train time (sec): " << (total_elapsed - timer_start) << "\n";


    // create the wrappers for all algorithms
    IKNWrapper* kn_wrapper = new IKNWrapper(moments, kney, qkney, order);

    PLREWrapper* plre_wrapper = NULL;

    if (order == 4)
        plre_wrapper = new PLREWrapper(moments, bi_plre, tri_plre, quad_plre);
    else if (order == 3)
        plre_wrapper = new PLREWrapper(moments, bi_plre, tri_plre);
    else if (order == 2)
        plre_wrapper = new PLREWrapper(moments, bi_plre);
    else
        assert(0);


    cout << "\n\n----------Saving model to archive----------\n";
    std::ofstream ofs(model_fn.c_str(), std::ios::binary);
    {
        boost::archive::binary_oarchive aro(ofs);

        // Write data
        aro << moments;
        aro << kn_wrapper;
        aro << plre_wrapper;
    }
    cout << "----------Done saving model to archive----------\n";

    delete(moments);
    delete(kn_wrapper);
    delete(plre_wrapper);

    if (test_fn.size() > 0)
    {
        // test archive
        std::ifstream ifs(model_fn.c_str(), std::ios::binary);
        boost::archive::binary_iarchive ari(ifs);
        SpecMoments* moments_ar;
        IKNWrapper* kn_ar;
        PLREWrapper* plre_ar;

        cout << "\n\n----------Loading model from archive----------\n";
        ari  & moments_ar;
        cout << "moments loaded\n";
        ari & kn_ar;
        cout << "KN loaded\n";
        ari & plre_ar;
        cout << "PLRE loaded\n";
        cout << "----------Done loading model from archive----------\n";

        // test
        cout << "\n\n----------Starting Test----------" << "\n";
        cout.flush();

        vector< vector<int>* > test_data;
        GetBigTestMatrix(test_data, test_fn, *moments_ar);

        QuadLMWrapper::RunQuadKneserNeyPLRE(*kn_ar, *plre_ar,
            *moments_ar, test_data, false);

        cout.flush();
        for (int i = 0; i < test_data.size(); ++i)
            delete(test_data[i]);

        delete(moments_ar);
        delete(kn_ar);
        delete(plre_ar);
    }

    return 0;
}
